
/*
 * moblin-app-installer -- Moblin garage client application.
 * Copyright (c) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Author: Robert Staudinger <robertx.staudinger@intel.com>
 */

#include <clutter-gtk/clutter-gtk.h>
#include <gtk/gtk.h>

#include "mai-application.h"
#include "mai-model.h"
#include "mai-shell.h"
#include "mai-transaction-manager.h"
#include "config.h"

G_DEFINE_TYPE (MaiApplication, mai_application, G_TYPE_OBJECT)

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), MAI_TYPE_APPLICATION, MaiApplicationPrivate))

enum
{
  PROP_0,

  PROP_WINDOW_GEOMETRY,
  PROP_APP_LIST
};

typedef struct
{
  gchar                 *window_geometry;
  gchar                 *app_list;
  MaiModel              *model;
  MaiTransactionManager *manager;

  GtkWidget             *window;
  GtkWidget             *embed;
  ClutterActor          *shell;
  ClutterActor          *stage;
} MaiApplicationPrivate;

static gboolean
mai_application_shutdown (MaiApplication  *self)
{
  /* TODO check if ready for shutdown. */
  clutter_main_quit ();
  return TRUE;
}

static void
_shell_show_uri_cb (MaiShell        *shell,
                    const gchar     *uri,
                    MaiApplication  *self)
{
  GError *error = NULL;

  gtk_show_uri (NULL, uri, clutter_get_current_event_time (), &error);
  if (error)
  {
    g_warning ("%s : %s", G_STRLOC, error->message);
    g_clear_error (&error);
  }
}

static gboolean
_window_delete_cb (GtkWidget      *widget,
                   GdkEvent       *event,
                   MaiApplication *self)
{
  MaiApplicationPrivate *priv = GET_PRIVATE (self);

  if (mai_application_shutdown (self))
  {
    priv->window = NULL;
    return FALSE;
  }

  /* Stop shutdown, do not delete window. */
  return TRUE;
}

static void
_shell_quit_request_cb (MaiShell        *shell,
                        MaiApplication  *self)
{
  mai_application_shutdown (self);
}

static void
_shell_cancel_cb (MaiShell        *shell,
                  const gchar     *package_name,
                  MaiApplication  *self)
{
  MaiApplicationPrivate *priv = GET_PRIVATE (self);
  GError                *error = NULL;

  mai_transaction_manager_cancel_by_name (priv->manager,
                                          package_name,
                                          &error);
  if (error)
  {
    g_warning ("%s : %s", G_STRLOC, error->message);
    g_clear_error (&error);
  }
}

static void
_shell_install_cb (MaiShell        *shell,
                   const gchar     *package_name,
                   MaiApplication  *self)
{
  MaiApplicationPrivate *priv = GET_PRIVATE (self);
  MaiPackage            *package;
  GError                *error = NULL;

  package = mai_model_lookup_package (priv->model, package_name);
  g_return_if_fail (package);

  mai_transaction_manager_queue (priv->manager,
                                 package,
                                 MAI_PACKAGE_STATUS_INSTALLING,
                                 &error);
  if (error)
  {
    g_warning ("%s : %s", G_STRLOC, error->message);
    g_clear_error (&error);
  }
}

static void
_shell_remove_cb (MaiShell        *shell,
                  const gchar     *package_name,
                  MaiApplication  *self)
{
  MaiApplicationPrivate *priv = GET_PRIVATE (self);
  MaiPackage            *package;
  GError                *error = NULL;

  package = mai_model_lookup_package (priv->model, package_name);
  g_return_if_fail (package);

  mai_transaction_manager_queue (priv->manager,
                                 package,
                                 MAI_PACKAGE_STATUS_REMOVING,
                                 &error);
  if (error)
  {
    g_warning ("%s : %s", G_STRLOC, error->message);
    g_clear_error (&error);
  }
}

static void
_shell_resolve_cb (MaiShell        *shell,
                   const gchar     *package_name,
                   MaiApplication  *self)
{
  MaiApplicationPrivate *priv = GET_PRIVATE (self);
  MaiPackage            *package;
  GError                *error = NULL;

  package = mai_model_lookup_package (priv->model, package_name);
  g_return_if_fail (package);

  mai_transaction_manager_queue (priv->manager,
                                 package,
                                 MAI_PACKAGE_STATUS_RESOLVING,
                                 &error);
  if (error)
  {
    g_warning ("%s : %s", G_STRLOC, error->message);
    g_clear_error (&error);
  }
}

static gboolean
_embed_configure_cb (GtkWidget          *widget,
                     GdkEventConfigure  *event,
                     MaiApplication     *self)
{
  MaiApplicationPrivate *priv = GET_PRIVATE (self);

  if (priv->shell)
    clutter_actor_set_size (priv->shell,
                            widget->allocation.width,
                            widget->allocation.height);

  return FALSE;
}

static void
_get_property (GObject    *object,
               guint       property_id,
               GValue     *value,
               GParamSpec *pspec)
{
  MaiApplicationPrivate *priv = GET_PRIVATE (object);

  switch (property_id)
  {
  case PROP_WINDOW_GEOMETRY:
    g_value_set_string (value, priv->window_geometry);
    break;
  case PROP_APP_LIST:
    g_value_set_string (value, priv->app_list);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_set_property (GObject      *object,
               guint         property_id,
               const GValue *value,
               GParamSpec   *pspec)
{
  MaiApplicationPrivate *priv = GET_PRIVATE (object);

  switch (property_id)
  {
  case PROP_WINDOW_GEOMETRY:
    /* Construct-only, no need for notifications and stuff. */
    g_return_if_fail (priv->window_geometry == NULL);
    priv->window_geometry = g_value_dup_string (value);
    break;
  case PROP_APP_LIST:
    /* Construct-only, no need for notifications and stuff. */
    g_return_if_fail (priv->app_list == NULL);
    priv->app_list = g_value_dup_string (value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_dispose (GObject *object)
{
  MaiApplicationPrivate *priv = GET_PRIVATE (object);

  if (priv->window_geometry)
  {
    g_free (priv->window_geometry);
    priv->window_geometry = NULL;
  }

  if (priv->app_list)
  {
    g_free (priv->app_list);
    priv->app_list = NULL;
  }

  if (priv->model)
  {
    g_object_unref (priv->model);
    priv->model = NULL;
  }

  if (priv->manager)
  {
    g_object_unref (priv->manager);
    priv->manager = NULL;
  }

  if (priv->window)
  {
    gtk_widget_destroy (priv->window);
    priv->window = NULL;
  }

  G_OBJECT_CLASS (mai_application_parent_class)->dispose (object);
}

static GObject *
_constructor (GType                  type,
              guint                  n_properties,
              GObjectConstructParam *properties)
{
  MaiApplication *self = (MaiApplication *)
                            G_OBJECT_CLASS (mai_application_parent_class)
                              ->constructor (type, n_properties, properties);
  MaiApplicationPrivate *priv = GET_PRIVATE (self);

  /* Model */
  priv->model = mai_model_new (priv->app_list);
  if (!priv->model)
    g_critical ("Failed to parse appstore data (%s)!", priv->app_list);

  /* Manager */
  priv->manager = mai_transaction_manager_new ();

  /* Window */
  priv->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_icon_name (GTK_WINDOW (priv->window), "system-software-install");
  g_signal_connect (priv->window, "delete-event",
                    G_CALLBACK (_window_delete_cb), self);

  priv->embed = gtk_clutter_embed_new ();
  g_signal_connect (priv->embed, "configure-event",
                    G_CALLBACK (_embed_configure_cb), self);
  gtk_container_add (GTK_CONTAINER (priv->window), priv->embed);

  if (priv->window_geometry)
  {
    gtk_window_parse_geometry (GTK_WINDOW (priv->window), priv->window_geometry);
  } else {
    GdkScreen    *screen = gdk_screen_get_default ();
    gint          monitor;
    GdkRectangle  rect;

    gtk_widget_realize (priv->window);
    monitor = gdk_screen_get_monitor_at_window (screen, priv->window->window);

    gdk_screen_get_monitor_geometry (screen, monitor, &rect);
    gtk_window_set_default_size (GTK_WINDOW (priv->window), rect.width, rect.height);
    gtk_window_set_decorated (GTK_WINDOW (priv->window), FALSE);
  }

  /* Stage */
  priv->stage = gtk_clutter_embed_get_stage (GTK_CLUTTER_EMBED (priv->embed));

  /* Shell */
  priv->shell = mai_shell_new (priv->model);
  clutter_container_add_actor (CLUTTER_CONTAINER (priv->stage), priv->shell);
  g_signal_connect (priv->shell, "show_uri",
                    G_CALLBACK (_shell_show_uri_cb), self);
  g_signal_connect (priv->shell, "quit-request",
                    G_CALLBACK (_shell_quit_request_cb), self);
  g_signal_connect (priv->shell, "cancel",
                    G_CALLBACK (_shell_cancel_cb), self);
  g_signal_connect (priv->shell, "install",
                    G_CALLBACK (_shell_install_cb), self);
  g_signal_connect (priv->shell, "remove",
                    G_CALLBACK (_shell_remove_cb), self);
  g_signal_connect (priv->shell, "resolve",
                    G_CALLBACK (_shell_resolve_cb), self);

  clutter_actor_show_all (priv->stage);
  gtk_widget_show_all (priv->window);

  return G_OBJECT (self);
}

static void
mai_application_class_init (MaiApplicationClass *klass)
{
  GObjectClass  *object_class = G_OBJECT_CLASS (klass);
  GParamFlags    param_flags;

  g_type_class_add_private (klass, sizeof (MaiApplicationPrivate));

  object_class->constructor = _constructor;
  object_class->get_property = _get_property;
  object_class->set_property = _set_property;
  object_class->dispose = _dispose;

  param_flags = G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY;

  g_object_class_install_property (object_class,
                                   PROP_WINDOW_GEOMETRY,
                                   g_param_spec_string ("window-geometry",
                                                        "Window geometry",
                                                        "Main window geometry",
                                                        NULL,
                                                        param_flags));
  g_object_class_install_property (object_class,
                                   PROP_APP_LIST,
                                   g_param_spec_string ("app-list",
                                                        "Application list",
                                                        "Application list file",
                                                        NULL,
                                                        param_flags));
}

static void
mai_application_init (MaiApplication *self)
{
}

MaiApplication *
mai_application_new (const gchar *app_list)
{
  return g_object_new (MAI_TYPE_APPLICATION,
                       "app-list", app_list,
                       NULL);
}

void
mai_application_run (MaiApplication *self)
{
  clutter_main ();
}

