
/*
 * moblin-app-installer -- Moblin garage client application.
 * Copyright (c) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Author: Robert Staudinger <robertx.staudinger@intel.com>
 */

#include <glib/gi18n.h>
#include "mai-category-button.h"
#include "mai-shell-dimensions.h"
#include "config.h"

static void _stylable_iface_init (NbtkStylableIface *iface);

G_DEFINE_TYPE_WITH_CODE (MaiCategoryButton, mai_category_button, NBTK_TYPE_BUTTON,
                         G_IMPLEMENT_INTERFACE (NBTK_TYPE_STYLABLE, _stylable_iface_init))

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), MAI_TYPE_CATEGORY_BUTTON, MaiCategoryButtonPrivate))

enum
{
  PROP_0,

  PROP_CATEGORY
};

typedef struct {
  /* Properties */
  MaiCategory const *category;

  /* Actors */
  ClutterActor  *icon;
  ClutterActor  *label;
} MaiCategoryButtonPrivate;

static void
mai_category_button_set_category (MaiCategoryButton *self,
                                  MaiCategory const *category);

/*
 * NbtkStylable, needed to get parent type styling applied.
 */

static const gchar *
_stylable_get_style_type (NbtkStylable *stylable)
{
  return "NbtkButton";
}

static void
_stylable_iface_init (NbtkStylableIface *iface)
{
  static gboolean _is_initialized = FALSE;

  if (!_is_initialized)
  {
    iface->get_style_type = _stylable_get_style_type;
    _is_initialized = TRUE;
  }
}

/*
 * MaiCategoryButton
 */

static void
_get_property (GObject    *object,
               guint       property_id,
               GValue     *value,
               GParamSpec *pspec)
{
  switch (property_id)
  {
  case PROP_CATEGORY:
    g_value_set_pointer (value,
                         (gpointer)
                            mai_category_button_get_category (
                              MAI_CATEGORY_BUTTON (object)));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_set_property (GObject      *object,
               guint         property_id,
               const GValue *value,
               GParamSpec   *pspec)
{
  switch (property_id)
  {
  case PROP_CATEGORY:
    mai_category_button_set_category (MAI_CATEGORY_BUTTON (object),
                                      g_value_get_pointer (value));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_dispose (GObject *object)
{
  /* MaiCategoryButtonPrivate *priv = GET_PRIVATE (object); */

  G_OBJECT_CLASS (mai_category_button_parent_class)->dispose (object);
}

static GObject *
_constructor (GType                  type,
              guint                  n_properties,
              GObjectConstructParam *properties)
{
  MaiCategoryButton *self = (MaiCategoryButton *)
                      G_OBJECT_CLASS (mai_category_button_parent_class)
                        ->constructor (type, n_properties, properties);

  nbtk_button_set_toggle_mode (NBTK_BUTTON (self), TRUE);

  return G_OBJECT (self);
}

static void
_get_preferred_width (ClutterActor  *self,
                      gfloat         for_height,
                      gfloat        *min_width_p,
                      gfloat        *natural_width_p)
{
  NbtkPadding    padding;
  gfloat         min_width = 0.;
  gfloat         natural_width = 0.;

  nbtk_widget_get_padding (NBTK_WIDGET (self), &padding);

  clutter_actor_get_preferred_width (nbtk_bin_get_child (NBTK_BIN (self)),
                                     for_height,
                                     &min_width,
                                     &natural_width);

  /* There seem to be serious rounding errors involved in measuring the text.
   * Maybe it's measuring before the big bold font is applied?
   * Hack around that for now. */

  if (min_width_p)
    *min_width_p = min_width + padding.left + padding.right + 32;

  if (natural_width_p)
    *natural_width_p = natural_width + padding.left + padding.right + 32;
}

static void
mai_category_button_class_init (MaiCategoryButtonClass *klass)
{
  GObjectClass      *object_class = G_OBJECT_CLASS (klass);
  ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
  GParamFlags        param_flags;

  g_type_class_add_private (klass, sizeof (MaiCategoryButtonPrivate));

  object_class->constructor = _constructor;
  object_class->get_property = _get_property;
  object_class->set_property = _set_property;
  object_class->dispose = _dispose;

  actor_class->get_preferred_width = _get_preferred_width;

  param_flags = G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_NAME;

  g_object_class_install_property (object_class,
                                   PROP_CATEGORY,
                                   g_param_spec_pointer ("category",
                                                         "Category",
                                                         "Package category",
                                                         param_flags));
}

static void
mai_category_button_init (MaiCategoryButton *self)
{
  MaiCategoryButtonPrivate  *priv;
  ClutterActor              *table;

  g_return_if_fail (MAI_IS_CATEGORY_BUTTON (self));

  priv = GET_PRIVATE (self);
  table = CLUTTER_ACTOR (nbtk_table_new ());
  nbtk_table_set_col_spacing (NBTK_TABLE (table), 10);

  priv->icon = clutter_texture_new ();
  clutter_actor_set_size (priv->icon,
                          MAI_CATEGORY_ICON_SIZE,
                          MAI_CATEGORY_ICON_SIZE);
  nbtk_table_add_actor (NBTK_TABLE (table), priv->icon, 0, 0);

  priv->label = CLUTTER_ACTOR (nbtk_label_new (""));
  nbtk_table_add_actor_with_properties (NBTK_TABLE (table), priv->label,
                                        0, 1,
                                        "x-fill",   FALSE,
                                        "y-fill",   FALSE,
                                        "x-expand", FALSE,
                                        "y-expand", FALSE,
                                        "x-align",  0.0,
                                        "y-align",  0.5,
                                        NULL);

  nbtk_bin_set_child (NBTK_BIN (self), table);
}

ClutterActor *
mai_category_button_new (MaiCategory const *category)
{
  return g_object_new (MAI_TYPE_CATEGORY_BUTTON,
                       "category", category,
                       NULL);
}

MaiCategory const *
mai_category_button_get_category (MaiCategoryButton const *self)
{
  g_return_val_if_fail (MAI_IS_CATEGORY_BUTTON (self), NULL);

  return GET_PRIVATE (self)->category;
}

static void
mai_category_button_set_category (MaiCategoryButton *self,
                                  MaiCategory const *category)
{
  MaiCategoryButtonPrivate  *priv;

  g_return_if_fail (MAI_IS_CATEGORY_BUTTON (self));
  g_return_if_fail (category);

  priv = GET_PRIVATE (self);

  /* Can only be set at construction. */
  g_return_if_fail (priv->category == NULL);

  priv->category = category;

  /* FIXME better fallback. */
  if (mai_category_get_icon (category))
  {
    gchar *filename = g_build_filename (ICONDIR,
                                        mai_category_get_icon (category),
                                        NULL);
    GError *error = NULL;
    clutter_texture_set_from_file (CLUTTER_TEXTURE (priv->icon),
                                   filename,
                                   &error);
    if (error)
      g_warning ("%s : %s", G_STRLOC, error->message), g_clear_error (&error);
  }

  nbtk_label_set_text (NBTK_LABEL (priv->label),
                       mai_category_get_name (priv->category));

  g_object_notify (G_OBJECT (self), "category");
}

