
/*
 * moblin-app-installer -- Moblin garage client application.
 * Copyright (c) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Author: Robert Staudinger <robertx.staudinger@intel.com>
 * Based on nbtk-expander.c by
 *    Copyright (c) 2009, Intel Corporation, by
 *    Thomas Wood <thomas.wood@intel.com>
 */

#include <glib/gi18n.h>
#include "mai-expander.h"
#include "mai-package-button.h"
#include "mai-shell-dimensions.h"

G_DEFINE_TYPE (MaiExpander, mai_expander, NBTK_TYPE_BIN)

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), MAI_TYPE_EXPANDER, MaiExpanderPrivate))

enum
{
  PROP_0,

  PROP_ICON,
  PROP_NAME,
  PROP_DESCRIPTION,
  PROP_STATUS,
  PROP_PROGRESS,
  PROP_STATUS_MESSAGE,
  PROP_PACKAGE_NAME,
  PROP_LONG_DESCRIPTION,

  PROP_EXPANDED
};

enum
{
  CANCEL,
  INSTALL,
  REMOVE,
  RESOLVE,

  LAST_SIGNAL
};

typedef struct
{
  /* Properties */
  gchar             *icon;
  MaiPackageStatus   status;
  gint               progress;
  gchar             *status_message;
  gchar             *package_name;
  gboolean           expanded;

  /* Actors */
  ClutterActor    *table;
  ClutterActor    *image;
  ClutterActor    *name;
  ClutterActor    *button;
  ClutterActor    *description;
  ClutterActor    *message;
  ClutterActor    *long_description;

  /* Data */
  ClutterTimeline *timeline;
  ClutterAlpha    *alpha;
  gdouble          animation_progress;
  gboolean         hovered;
} MaiExpanderPrivate;

static guint _signals[LAST_SIGNAL] = { 0, };

static gchar *
format_message (const gchar   *label,
                gint           progress)
{
  if (progress >= 0 && progress <= 100)
  {
    return g_strdup_printf ("%s %d%c",
                            label,
                            progress,
                            '%');
  } else {
    return g_strdup (label);
  }
}

static void
update_status (MaiExpander *self)
{
  MaiExpanderPrivate *priv = GET_PRIVATE (self);
  gchar *label;

  switch (priv->status)
  {
  case MAI_PACKAGE_STATUS_QUEUED_FOR_RESOLVE:
    clutter_actor_set_reactive (priv->button, TRUE);
    nbtk_button_set_label (NBTK_BUTTON (priv->button), _("Cancel"));
    label = format_message (_("Queued for resolve ..."), priv->progress);
    nbtk_label_set_text (NBTK_LABEL (priv->message), label);
    g_free (label);
    break;
  case MAI_PACKAGE_STATUS_RESOLVING:
    /* Resolve can't be cancelled. */
    clutter_actor_set_reactive (priv->button, FALSE);
    nbtk_button_set_label (NBTK_BUTTON (priv->button), "");
    label = format_message (_("Resolving ..."), priv->progress);
    nbtk_label_set_text (NBTK_LABEL (priv->message), label);
    g_free (label);
    break;
  case MAI_PACKAGE_STATUS_AVAILABLE:
    clutter_actor_set_reactive (priv->button, TRUE);
    nbtk_button_set_label (NBTK_BUTTON (priv->button), _("Install"));
    nbtk_label_set_text (NBTK_LABEL (priv->message), _("Available"));
    break;
  case MAI_PACKAGE_STATUS_QUEUED_FOR_INSTALLATION:
    clutter_actor_set_reactive (priv->button, TRUE);
    nbtk_button_set_label (NBTK_BUTTON (priv->button), _("Cancel"));
    label = format_message (_("Queued for installation ..."), priv->progress);
    nbtk_label_set_text (NBTK_LABEL (priv->message), label);
    g_free (label);
    break;
  case MAI_PACKAGE_STATUS_INSTALLING:
    /* Installation can't be cancelled. */
    clutter_actor_set_reactive (priv->button, FALSE);
    nbtk_button_set_label (NBTK_BUTTON (priv->button), "");
    label = format_message (_("Installing ..."), priv->progress);
    nbtk_label_set_text (NBTK_LABEL (priv->message), label);
    g_free (label);
    break;
  case MAI_PACKAGE_STATUS_INSTALLED:
    clutter_actor_set_reactive (priv->button, TRUE);
    nbtk_button_set_label (NBTK_BUTTON (priv->button), _("Remove"));
    nbtk_label_set_text (NBTK_LABEL (priv->message), _("Installed"));
    break;
  case MAI_PACKAGE_STATUS_QUEUED_FOR_REMOVAL:
    clutter_actor_set_reactive (priv->button, TRUE);
    nbtk_button_set_label (NBTK_BUTTON (priv->button), _("Cancel"));
    label = format_message (_("Queued for removal ..."), priv->progress);
    nbtk_label_set_text (NBTK_LABEL (priv->message), label);
    g_free (label);
    break;
  case MAI_PACKAGE_STATUS_REMOVING:
    /* Removal can't be cancelled. */
    clutter_actor_set_reactive (priv->button, FALSE);
    nbtk_button_set_label (NBTK_BUTTON (priv->button), "");
    label = format_message (_("Removing ..."), priv->progress);
    nbtk_label_set_text (NBTK_LABEL (priv->message), label);
    g_free (label);
    break;
  case MAI_PACKAGE_STATUS_ERROR:
    clutter_actor_set_reactive (priv->button, TRUE);
    nbtk_button_set_label (NBTK_BUTTON (priv->button), _("Retry"));
    if (priv->status_message)
       nbtk_label_set_text (NBTK_LABEL (priv->message), priv->status_message);
    else
       nbtk_label_set_text (NBTK_LABEL (priv->message), _("Unknown error"));
    break;
  }
}

static void
_button_clicked_cb (NbtkButton  *button,
                    MaiExpander *self)
{
  MaiExpanderPrivate *priv = GET_PRIVATE (self);

  /* Filter accidential double-clicks.
   * Button is re-enabled in update_status(). */
  clutter_actor_set_reactive (priv->button, FALSE);

  switch (priv->status)
  {
  case MAI_PACKAGE_STATUS_QUEUED_FOR_RESOLVE:
    g_signal_emit_by_name (self, "cancel");
    break;
  case MAI_PACKAGE_STATUS_RESOLVING:
    g_signal_emit_by_name (self, "cancel");
    break;
  case MAI_PACKAGE_STATUS_AVAILABLE:
    g_signal_emit_by_name (self, "install");
    break;
  case MAI_PACKAGE_STATUS_QUEUED_FOR_INSTALLATION:
    g_signal_emit_by_name (self, "cancel");
    break;
  case MAI_PACKAGE_STATUS_INSTALLING:
    g_signal_emit_by_name (self, "cancel");
    break;
  case MAI_PACKAGE_STATUS_INSTALLED:
    g_signal_emit_by_name (self, "remove");
    break;
  case MAI_PACKAGE_STATUS_QUEUED_FOR_REMOVAL:
    g_signal_emit_by_name (self, "cancel");
    break;
  case MAI_PACKAGE_STATUS_REMOVING:
    g_signal_emit_by_name (self, "cancel");
    break;
  case MAI_PACKAGE_STATUS_ERROR:
    g_signal_emit_by_name (self, "resolve");
    break;
  }
}

static void
_get_property (GObject    *object,
               guint       property_id,
               GValue     *value,
               GParamSpec *pspec)
{
  MaiExpanderPrivate *priv = GET_PRIVATE (object);

  switch (property_id)
  {
    case PROP_ICON:
      g_value_set_string (value, priv->icon);
      break;
    case PROP_NAME:
      g_value_set_string (value,
                          clutter_text_get_text (CLUTTER_TEXT (priv->name)));
      break;
    case PROP_DESCRIPTION:
      g_value_set_string (value,
                          nbtk_label_get_text (NBTK_LABEL (priv->description)));
      break;
    case PROP_STATUS:
      g_value_set_uint (value,
                        mai_expander_get_status (MAI_EXPANDER (object)));
      break;
    case PROP_PROGRESS:
      g_value_set_int (value,
                       mai_expander_get_progress (MAI_EXPANDER (object)));
      break;
    case PROP_STATUS_MESSAGE:
      g_value_set_string (value,
                          mai_expander_get_status_message (MAI_EXPANDER (object)));
      break;
    case PROP_PACKAGE_NAME:
      g_value_set_string (value,
                          mai_expander_get_package_name (MAI_EXPANDER (object)));
      break;
    case PROP_LONG_DESCRIPTION:
      g_value_set_string (value,
                          mai_expander_get_long_description (MAI_EXPANDER (object)));
      break;
    case PROP_EXPANDED:
      g_value_set_boolean (value, priv->expanded);
      break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_set_property (GObject      *object,
               guint         property_id,
               const GValue *value,
               GParamSpec   *pspec)
{
  MaiExpanderPrivate *priv = GET_PRIVATE (object);

  switch (property_id)
  {
    case PROP_ICON:
      /* Construct-only, no need to notify and stuff. */
      priv->icon = g_value_dup_string (value);
      if (priv->icon)
      {
        GError *error = NULL;
        if (g_path_is_absolute (priv->icon))
        {
          clutter_texture_set_from_file (CLUTTER_TEXTURE (priv->image),
                                         priv->icon, &error);
        } else {
          gchar *path = g_build_filename (APPICONDIR,
                                          priv->icon,
                                          NULL);
          clutter_texture_set_from_file (CLUTTER_TEXTURE (priv->image),
                                         path, &error);
          g_free (path);
        }
        if (error)
        {
          g_warning ("%s : %s", G_STRLOC, error->message), g_clear_error (&error);
        } else {
          gint width, height;
          clutter_texture_get_base_size (CLUTTER_TEXTURE (priv->image),
                                         &width, &height);
          clutter_actor_set_size (priv->image,
                                  MAI_APPLICATION_ICON_SIZE,
                                  MAI_APPLICATION_ICON_SIZE);
        }
      }
      break;
    case PROP_NAME:
      /* Construct-only, no need to notify and stuff. */
      clutter_text_set_text (CLUTTER_TEXT (priv->name),
                             g_value_get_string (value));
      break;
    case PROP_DESCRIPTION:
      /* Construct-only, no need to notify and stuff. */
      nbtk_label_set_text (NBTK_LABEL (priv->description),
                           g_value_get_string (value));
      break;
    case PROP_STATUS:
      mai_expander_set_status (MAI_EXPANDER (object),
                               g_value_get_uint (value));
      break;
    case PROP_PROGRESS:
      mai_expander_set_progress (MAI_EXPANDER (object),
                                 g_value_get_int (value));
      break;
    case PROP_STATUS_MESSAGE:
      mai_expander_set_status_message (MAI_EXPANDER (object),
                                       g_value_get_string (value));
      break;
    case PROP_PACKAGE_NAME:
      /* Construct-only, no need to notify and stuff. */
      priv->package_name = g_value_dup_string (value);
      break;
    case PROP_LONG_DESCRIPTION:
      /* Construct-only, no need to notify and stuff. */
      nbtk_label_set_text (NBTK_LABEL (priv->long_description),
                           g_value_get_string (value));
      break;
    case PROP_EXPANDED:
      mai_expander_set_expanded (MAI_EXPANDER (object),
                                 g_value_get_boolean (value));
      break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_dispose (GObject *object)
{
  MaiExpanderPrivate *priv = GET_PRIVATE (object);

  if (priv->timeline)
  {
    g_object_unref (priv->timeline);
    priv->timeline = NULL;
  }

  if (priv->alpha)
  {
    g_object_unref (priv->alpha);
    priv->alpha = NULL;
  }

  G_OBJECT_CLASS (mai_expander_parent_class)->dispose (object);
}

static void
_finalize (GObject *object)
{
  MaiExpanderPrivate *priv = GET_PRIVATE (object);

  g_free (priv->icon);
  priv->icon = NULL;

  g_free (priv->status_message);
  priv->status_message = NULL;

  g_free (priv->package_name);
  priv->package_name = NULL;

  G_OBJECT_CLASS (mai_expander_parent_class)->finalize (object);
}

static void
timeline_complete (ClutterTimeline *timeline,
                   ClutterActor    *expander)
{
  MaiExpanderPrivate *priv = GET_PRIVATE (expander);
  guchar opacity;

  if (priv->expanded)
  {
    opacity = clutter_actor_get_opacity (priv->long_description);
    clutter_actor_set_opacity (priv->long_description, 0);

    clutter_actor_show (priv->long_description);
    clutter_actor_animate (priv->long_description, CLUTTER_EASE_IN_SINE, 100,
                           "opacity", opacity,
                           NULL);
  } else {
    clutter_actor_hide (priv->long_description);
  }

}

static void
new_frame (ClutterTimeline *timeline,
           gint             frame_num,
           ClutterActor    *expander)
{
  MaiExpanderPrivate *priv = GET_PRIVATE (expander);

  priv->animation_progress = clutter_alpha_get_alpha (priv->alpha);

  clutter_actor_queue_relayout (expander);
}

static void
mai_expander_toggle_expanded (MaiExpander *expander)
{
  MaiExpanderPrivate *priv = GET_PRIVATE (expander);

  priv->expanded = !priv->expanded;
  g_object_notify ((GObject *) expander, "expanded");

  if (priv->expanded)
    {
      nbtk_widget_set_style_pseudo_class (NBTK_WIDGET (expander), "active");
    }
  else
    {
      clutter_actor_hide (priv->long_description);
      if (priv->hovered)
        nbtk_widget_set_style_pseudo_class (NBTK_WIDGET (expander), "hover");
      else
        nbtk_widget_set_style_pseudo_class (NBTK_WIDGET (expander), NULL);
    }

  /* setup and start the expansion animation */
  if (!priv->expanded)
    {
      clutter_timeline_set_direction (priv->timeline,
                                      CLUTTER_TIMELINE_BACKWARD);
    }
  else
    {
      clutter_timeline_set_direction (priv->timeline,
                                      CLUTTER_TIMELINE_FORWARD);
    }

  if (!clutter_timeline_is_playing (priv->timeline))
    clutter_timeline_rewind (priv->timeline);

  clutter_timeline_start (priv->timeline);
}

static gboolean
_button_release_event (ClutterActor       *actor,
                       ClutterButtonEvent *event)
{
  mai_expander_toggle_expanded (MAI_EXPANDER (actor));

  return FALSE;
}

static void
mai_expander_style_changed (NbtkWidget *widget)
{
  MaiExpanderPrivate *priv = GET_PRIVATE (widget);
  const gchar *pseudo_class;
  ClutterColor *color = NULL;
  gchar *font_name;
  gchar *font_string;
  gint font_size;

  pseudo_class = nbtk_widget_get_style_pseudo_class (widget);

  nbtk_stylable_get (NBTK_STYLABLE (widget),
                     "color", &color,
                     "font-family", &font_name,
                     "font-size", &font_size,
                     NULL);

  if (color)
    {
      clutter_text_set_color (CLUTTER_TEXT (priv->name), color);
      clutter_color_free (color);
    }

  if (font_name || font_size)
    {
      if (font_name && font_size)
        {
          font_string = g_strdup_printf ("%s %dpx", font_name, font_size);
          g_free (font_name);
        }
      else
        {
          if (font_size)
            font_string = g_strdup_printf ("%dpx", font_size);
          else
            font_string = font_name;
        }

      clutter_text_set_font_name (CLUTTER_TEXT (priv->name), font_string);
      g_free (font_string);
    }

}

static void
_get_preferred_height (ClutterActor *actor,
                       gfloat        for_width,
                       gfloat       *min_height_p,
                       gfloat       *natural_height_p)
{
  MaiExpanderPrivate *priv = GET_PRIVATE (actor);
  NbtkPadding         padding;

  nbtk_widget_get_padding (NBTK_WIDGET (actor), &padding);

  clutter_actor_get_preferred_height (priv->table,
                                      for_width,
                                      min_height_p,
                                      natural_height_p);

  if (min_height_p)
    *min_height_p += padding.top + padding.bottom;

  if (natural_height_p)
    *natural_height_p += padding.top + padding.bottom;

  if (clutter_timeline_is_playing (priv->timeline))
  {
    gfloat min_description_height;
    gfloat natural_description_height;

    clutter_actor_get_preferred_height (priv->long_description,
                                        for_width,
                                        &min_description_height,
                                        &natural_description_height);

    if (min_height_p)
      *min_height_p += priv->animation_progress * min_description_height;

    if (natural_height_p)
      *natural_height_p += priv->animation_progress * natural_description_height;
  }
}

static gboolean
_enter_event (ClutterActor         *actor,
              ClutterCrossingEvent *event)
{
  MaiExpanderPrivate *priv = GET_PRIVATE (actor);

  priv->hovered = TRUE;
  if (!priv->expanded)
    nbtk_widget_set_style_pseudo_class (NBTK_WIDGET (actor), "hover");

  return FALSE;
}

static gboolean
_leave_event (ClutterActor         *actor,
              ClutterCrossingEvent *event)
{
  MaiExpanderPrivate *priv = GET_PRIVATE (actor);

  priv->hovered = FALSE;
  if (!priv->expanded)
    nbtk_widget_set_style_pseudo_class (NBTK_WIDGET (actor), NULL);

  return FALSE;
}

static void
mai_expander_class_init (MaiExpanderClass *klass)
{
  GObjectClass      *object_class = G_OBJECT_CLASS (klass);
  ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
  GParamFlags        param_flags;

  g_type_class_add_private (klass, sizeof (MaiExpanderPrivate));

  object_class->get_property = _get_property;
  object_class->set_property = _set_property;
  object_class->dispose = _dispose;
  object_class->finalize = _finalize;

  actor_class->button_release_event = _button_release_event;
  actor_class->get_preferred_height = _get_preferred_height;
  actor_class->enter_event = _enter_event;
  actor_class->leave_event = _leave_event;

  /* Properties */

  param_flags = G_PARAM_READWRITE | G_PARAM_CONSTRUCT |
                G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB;

  g_object_class_install_property (object_class,
                                   PROP_ICON,
                                   g_param_spec_string ("icon", "Icon",
                                                        "Application icon",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_NAME,
                                   g_param_spec_string ("name", "Name",
                                                        "Application name",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_DESCRIPTION,
                                   g_param_spec_string ("description",
                                                        "Description",
                                                        "Application description",
                                                        "", param_flags));
  /* FIXME turn into an enum derived type. */
  g_object_class_install_property (object_class,
                                   PROP_STATUS,
                                   g_param_spec_uint ("status", "Status",
                                                      "Package status",
                                                      MAI_PACKAGE_STATUS_RESOLVING,
                                                      MAI_PACKAGE_STATUS_ERROR,
                                                      MAI_PACKAGE_STATUS_RESOLVING,
                                                      param_flags));
  g_object_class_install_property (object_class,
                                   PROP_PROGRESS,
                                   g_param_spec_int ("progress", "Progress",
                                                      "Transaction progress",
                                                      -1, 100, -1,
                                                      param_flags));
  g_object_class_install_property (object_class,
                                   PROP_STATUS_MESSAGE,
                                   g_param_spec_string ("status-message",
                                                        "Status message",
                                                        "Package status message",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_PACKAGE_NAME,
                                   g_param_spec_string ("package-name",
                                                        "Package name",
                                                        "Name of package",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_LONG_DESCRIPTION,
                                   g_param_spec_string ("long-description",
                                                        "Long description",
                                                        "Long package description",
                                                        "", param_flags));
  g_object_class_install_property (object_class,
                                   PROP_EXPANDED,
                                   g_param_spec_boolean ("expanded", "Expanded",
                                                         "Indicates if the expander is open or closed",
                                                         FALSE,
                                                         param_flags));

  /* Signals */

  _signals[CANCEL] = g_signal_new ("cancel",
                                   G_TYPE_FROM_CLASS (klass),
                                   G_SIGNAL_RUN_LAST,
                                   G_STRUCT_OFFSET (MaiExpanderClass, cancel),
                                   NULL, NULL,
                                   g_cclosure_marshal_VOID__VOID,
                                   G_TYPE_NONE, 0);

  _signals[INSTALL] = g_signal_new ("install",
                                    G_TYPE_FROM_CLASS (klass),
                                    G_SIGNAL_RUN_LAST,
                                    G_STRUCT_OFFSET (MaiExpanderClass, install),
                                    NULL, NULL,
                                    g_cclosure_marshal_VOID__VOID,
                                    G_TYPE_NONE, 0);

  _signals[REMOVE] = g_signal_new ("remove",
                                   G_TYPE_FROM_CLASS (klass),
                                   G_SIGNAL_RUN_LAST,
                                   G_STRUCT_OFFSET (MaiExpanderClass, remove),
                                   NULL, NULL,
                                   g_cclosure_marshal_VOID__VOID,
                                   G_TYPE_NONE, 0);

  _signals[RESOLVE] = g_signal_new ("resolve",
                                    G_TYPE_FROM_CLASS (klass),
                                    G_SIGNAL_RUN_LAST,
                                    G_STRUCT_OFFSET (MaiExpanderClass, resolve),
                                    NULL, NULL,
                                    g_cclosure_marshal_VOID__VOID,
                                    G_TYPE_NONE, 0);
}

static void
mai_expander_init (MaiExpander *self)
{
  MaiExpanderPrivate  *priv = GET_PRIVATE (self);
  ClutterText         *text;

  priv->table = CLUTTER_ACTOR (nbtk_table_new ());
  nbtk_table_set_col_spacing (NBTK_TABLE (priv->table), 10.);
  nbtk_bin_set_child (NBTK_BIN (self), priv->table);
  nbtk_bin_set_alignment (NBTK_BIN (self), NBTK_ALIGN_START, NBTK_ALIGN_START);
  nbtk_bin_set_fill (NBTK_BIN (self), TRUE, FALSE);

  priv->image = clutter_texture_new ();
  clutter_actor_set_name (priv->image, "image");
  clutter_actor_show (priv->image);
  nbtk_table_add_actor_with_properties (NBTK_TABLE (priv->table),
                                        priv->image,
                                        0, 0,
                                        "x-fill",   FALSE,
                                        "y-fill",   FALSE,
                                        "x-expand", FALSE,
                                        "y-expand", FALSE,
                                        "x-align",  0.0,
                                        "y-align",  0.0,
                                        "allocate-hidden", TRUE,
                                        "row-span", 3,
                                        NULL);

  priv->name = clutter_text_new ();
  nbtk_table_add_actor_with_properties (NBTK_TABLE (priv->table),
                                        priv->name,
                                        0, 1,
                                        "x-fill",   TRUE,
                                        "y-fill",   FALSE,
                                        "x-expand", TRUE,
                                        "y-expand", FALSE,
                                        "x-align",  0.0,
                                        "y-align",  0.0,
                                        "allocate-hidden", TRUE,
                                        NULL);

  priv->button = mai_package_button_new_with_label ("action");
  nbtk_table_add_actor_with_properties (NBTK_TABLE (priv->table),
                                        priv->button,
                                        0, 2,
                                        "x-fill",   FALSE,
                                        "y-fill",   FALSE,
                                        "x-expand", FALSE,
                                        "y-expand", FALSE,
                                        "x-align",  1.0,
                                        "y-align",  0.5,
                                        "allocate-hidden", TRUE,
                                        "row-span", 3,
                                        NULL);
  g_signal_connect (priv->button, "clicked",
                    G_CALLBACK (_button_clicked_cb), self);

  priv->description = CLUTTER_ACTOR (nbtk_label_new (""));
  nbtk_table_add_actor_with_properties (NBTK_TABLE (priv->table),
                                        priv->description,
                                        1, 1,
                                        "x-fill",   FALSE,
                                        "y-fill",   FALSE,
                                        "x-expand", FALSE,
                                        "y-expand", FALSE,
                                        "x-align",  0.0,
                                        "y-align",  0.0,
                                        "allocate-hidden", FALSE,
                                        NULL);

  priv->message = CLUTTER_ACTOR (nbtk_label_new (""));
  nbtk_widget_set_style_class_name (NBTK_WIDGET (priv->message),
                                    "message");
  nbtk_table_add_actor_with_properties (NBTK_TABLE (priv->table),
                                        priv->message,
                                        2, 1,
                                        "x-fill",   FALSE,
                                        "y-fill",   FALSE,
                                        "x-expand", FALSE,
                                        "y-expand", FALSE,
                                        "x-align",  0.0,
                                        "y-align",  0.0,
                                        "allocate-hidden", FALSE,
                                        NULL);

  text = CLUTTER_TEXT (nbtk_label_get_clutter_text (NBTK_LABEL (priv->message)));
  clutter_text_set_ellipsize (text, PANGO_ELLIPSIZE_NONE);
  clutter_text_set_line_wrap (text, TRUE);

  priv->long_description = CLUTTER_ACTOR (nbtk_label_new (""));
  nbtk_widget_set_style_class_name (NBTK_WIDGET (priv->long_description),
                                    "description");
  clutter_actor_hide (priv->long_description);
  nbtk_table_add_actor_with_properties (NBTK_TABLE (priv->table),
                                        priv->long_description,
                                        3, 0,
                                        "x-fill",   FALSE,
                                        "y-fill",   FALSE,
                                        "x-expand", FALSE,
                                        "y-expand", FALSE,
                                        "x-align",  0.0,
                                        "y-align",  0.0,
                                        "col-span", 3,
                                        "allocate-hidden", FALSE,
                                        NULL);

  text = CLUTTER_TEXT (nbtk_label_get_clutter_text (NBTK_LABEL (priv->long_description)));
  clutter_text_set_ellipsize (text, PANGO_ELLIPSIZE_NONE);
  clutter_text_set_justify (text, TRUE);
  clutter_text_set_line_wrap (text, TRUE);

  priv->timeline = clutter_timeline_new (250);
  g_signal_connect (priv->timeline, "new-frame", G_CALLBACK (new_frame), self);
  g_signal_connect (priv->timeline, "completed", G_CALLBACK (timeline_complete), self);

  priv->alpha = clutter_alpha_new_full (priv->timeline, CLUTTER_EASE_IN_SINE);
  g_object_ref_sink (priv->alpha);

  clutter_actor_set_reactive ((ClutterActor *) self, TRUE);

  g_signal_connect (self, "style-changed",
                    G_CALLBACK (mai_expander_style_changed), NULL);
}

/**
 * mai_expander_new:
 *
 * Creates a new #MaiExpander
 *
 * Returns: the newly allocated #MaiExpander
 */
NbtkWidget*
mai_expander_new (void)
{
  return g_object_new (MAI_TYPE_EXPANDER, NULL);
}

const gchar *
mai_expander_get_icon (MaiExpander const *self)
{
  g_return_val_if_fail (self, FALSE);

  return GET_PRIVATE (self)->icon;
}

const gchar *
mai_expander_get_name (MaiExpander const *self)
{
  g_return_val_if_fail (self, FALSE);

  return nbtk_label_get_text (NBTK_LABEL (GET_PRIVATE (self)->name));
}

const gchar *
mai_expander_get_description  (MaiExpander const *self)
{
  g_return_val_if_fail (self, FALSE);

  return nbtk_label_get_text (NBTK_LABEL (GET_PRIVATE (self)->description));
}

const gchar *
mai_expander_get_package_name (MaiExpander const *self)
{
  g_return_val_if_fail (self, FALSE);

  return GET_PRIVATE (self)->package_name;
}

MaiPackageStatus
mai_expander_get_status (MaiExpander const *self)
{
  g_return_val_if_fail (self, MAI_PACKAGE_STATUS_ERROR);

  return GET_PRIVATE (self)->status;
}

void
mai_expander_set_status (MaiExpander       *self,
                         MaiPackageStatus   status)
{
  MaiExpanderPrivate *priv;

  g_return_if_fail (MAI_IS_EXPANDER (self));

  priv = GET_PRIVATE (self);

  if (priv->status != status)
  {
    priv->status = status;
    g_object_notify (G_OBJECT (self), "status");
    update_status (self);
  }
}

gint
mai_expander_get_progress (MaiExpander const  *self)
{
  g_return_val_if_fail (self, -1);

  return GET_PRIVATE (self)->progress;
}

void
mai_expander_set_progress (MaiExpander *self,
                           gint         progress)
{
  MaiExpanderPrivate *priv;

  g_return_if_fail (MAI_IS_EXPANDER (self));

  priv = GET_PRIVATE (self);
  g_return_if_fail (progress >= -1);
  g_return_if_fail (progress <= 100);

  if (priv->progress != progress)
  {
    priv->progress = progress;
    g_object_notify (G_OBJECT (self), "progress");
    update_status (self);
  }
}

const gchar *
mai_expander_get_status_message (MaiExpander const  *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->status_message;
}

void
mai_expander_set_status_message (MaiExpander *self,
                                 const gchar *status_message)
{
  MaiExpanderPrivate *priv;

  g_return_if_fail (MAI_IS_EXPANDER (self));

  priv = GET_PRIVATE (self);

  if (0 != g_strcmp0 (priv->status_message, status_message))
  {
    g_free (priv->status_message);
    priv->status_message = g_strdup (status_message);
    g_object_notify (G_OBJECT (self), "status-message");
    update_status (self);
  }
}

const gchar *
mai_expander_get_long_description (MaiExpander const *self)
{
  g_return_val_if_fail (MAI_IS_EXPANDER (self), NULL);

  return nbtk_label_get_text (NBTK_LABEL (GET_PRIVATE (self)->long_description));
}

/**
 * mai_expander_set_expanded:
 * @expander: A #MaiExpander
 * @expanded: the state of the expander to set
 *
 * Set the state of the expander. This will cause the expander to open and
 * close if the state is changed.
 */
void
mai_expander_set_expanded (MaiExpander *expander,
                            gboolean      expanded)
{
  MaiExpanderPrivate *priv;

  g_return_if_fail (MAI_IS_EXPANDER (expander));

  priv = GET_PRIVATE (expander);

  if (priv->expanded != expanded)
    {
      mai_expander_toggle_expanded (expander);
    }
}

/**
 * mai_expander_get_expanded:
 * @expander: a #MaiExpander
 *
 * Get the current state of the expander
 *
 * Returns: #TRUE if the expander is open, #FALSE if it is closed
 */
gboolean
mai_expander_get_expanded (MaiExpander *expander)
{
  g_return_val_if_fail (MAI_IS_EXPANDER (expander), FALSE);

  return GET_PRIVATE (expander)->expanded;
}
