
/*
 * moblin-app-installer -- Moblin garage client application.
 * Copyright (c) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Author: Robert Staudinger <robertx.staudinger@intel.com>
 */

#include <glib/gi18n.h>
#include "mai-model.h"
#include "mai-model-priv.h"
#include "mai-parser.h"
#include "mai-resolver.h"
#include "config.h"

#if 0 /* unused */

#define MAI_MODEL_ERROR (mai_model_error_quark ())

static GQuark
mai_model_error_quark (void)
{
  static GQuark _quark = 0;
  if (!_quark)
    _quark = g_quark_from_static_string ("mai-model-error");
  return _quark;
}

#endif

G_DEFINE_TYPE (MaiModel, mai_model, G_TYPE_OBJECT)

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), MAI_TYPE_MODEL, MaiModelPrivate))

enum
{
  PROP_0,

  PROP_APPSTORE_FILE,
  PROP_STORE,
  PROP_CATEGORIES,
};

typedef struct {
  /* Properties */
  gchar         *appstore_file;
  ClutterModel  *store;
  GList         *categories;
} MaiModelPrivate;

static PkControl *_control = NULL;

PkControl *
mai_model_get_control (void)
{
  if (NULL == _control)
  {
    /* This reference is release in _finalize(). */
    _control = pk_control_new ();
    g_object_add_weak_pointer (G_OBJECT (_control), (gpointer *) &_control);
  }

  return _control;
}

static void
_get_property (GObject    *object,
               guint       property_id,
               GValue     *value,
               GParamSpec *pspec)
{
  switch (property_id) {
  case PROP_APPSTORE_FILE:
    g_value_set_string (value,
                        mai_model_get_appstore_file (MAI_MODEL (object)));
    break;
  case PROP_STORE:
    g_value_set_object (value,
                        mai_model_get_store (MAI_MODEL (object)));
    break;
  case PROP_CATEGORIES:
    g_value_set_pointer (value,
                         (gpointer) mai_model_get_categories (MAI_MODEL (object)));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_set_property (GObject      *object,
               guint         property_id,
               const GValue *value,
               GParamSpec   *pspec)
{
  MaiModelPrivate *priv = GET_PRIVATE (object);

  switch (property_id) {
  case PROP_APPSTORE_FILE:
    /* Construct-only, no need to notify and stuff. */
    priv->appstore_file = g_value_dup_string (value);
    break;
  case PROP_STORE:
    /* Construct-only, no need to notify and stuff. */
    if (priv->store)
      g_object_unref (priv->store), priv->store = NULL;
    if (g_value_get_object (value))
      priv->store = g_object_ref (g_value_get_object (value));
    break;
  case PROP_CATEGORIES:
    /* Read-only property. */
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_dispose (GObject *object)
{
  MaiModelPrivate *priv = GET_PRIVATE (object);

  if (priv->appstore_file)
    g_free (priv->appstore_file), priv->appstore_file = NULL;

  if (priv->store)
    g_object_unref (priv->store), priv->store = NULL;

  if (priv->categories)
  {
    g_list_foreach (priv->categories, (GFunc) g_object_unref, NULL);
    g_list_free (priv->categories);
    priv->categories = NULL;
  }

  if (_control)
  {
    g_object_unref (_control);
    /* Does the weak pointer work? */
    g_warn_if_fail (_control == NULL);
  }

  G_OBJECT_CLASS (mai_model_parent_class)->dispose (object);
}

static gboolean
_package_set_error_if_not_resolved (ClutterModel      *model,
                                    ClutterModelIter  *iter,
                                    gpointer           user_data)
{
  MaiPackage *package = NULL;
  clutter_model_iter_get (iter, 0, &package, -1);

  if (MAI_PACKAGE_STATUS_RESOLVING == mai_package_get_status (package))
  {
    mai_package_set_message (package, _("Package not found"));
    mai_package_set_status (package, MAI_PACKAGE_STATUS_ERROR);
  }

  return TRUE;
}

static void
_resolver_notify_progress_cb (MaiResolver  *resolver,
                              GParamSpec   *pspec,
                              MaiModel     *self)
{
  MaiModelPrivate *priv = GET_PRIVATE (self);

  if (-1 == mai_resolver_get_progress (resolver))
  {
    g_object_unref (resolver);

    mai_object_store_foreach_unfiltered (MAI_OBJECT_STORE (priv->store),
                                         _package_set_error_if_not_resolved,
                                         NULL);
  }
}

static GObject *
_constructor (GType                  type,
              guint                  n_properties,
              GObjectConstructParam *properties)
{
  MaiModel *self = (MaiModel *)
                      G_OBJECT_CLASS (mai_model_parent_class)
                          ->constructor (type, n_properties, properties);
  MaiModelPrivate     *priv = GET_PRIVATE (self);
  MaiParser           *parser = mai_parser_new (priv->appstore_file);
  MaiParserIterator   *iter = NULL;
  MaiResolver         *resolver = mai_resolver_new ();
  GError              *error = NULL;

  g_return_val_if_fail (priv->appstore_file, NULL);
  g_return_val_if_fail (MAI_IS_PARSER (parser), NULL);

  /* Build application store. */
  priv->store = mai_object_store_new (11,
                                      MAI_TYPE_PACKAGE, "package",
                                      G_TYPE_STRING,    "package-name",
                                      G_TYPE_STRING,    "name",
                                      G_TYPE_STRING,    "version",
                                      G_TYPE_STRING,    "description",
                                      G_TYPE_STRING,    "long-description",
                                      G_TYPE_STRING,    "icon",
                                      G_TYPE_POINTER,   "categories",
                                      G_TYPE_INT,       "status", /* FIXME enum */
                                      G_TYPE_INT,       "progress",
                                      G_TYPE_STRING,    "message");
  for (iter = mai_parser_get_package_iterator (parser);
       !mai_parser_iterator_is_end (iter);
       mai_parser_iterator_next (iter))
  {
    MaiPackage *package = mai_parser_iterator_create_package (iter);
    if (!package)
      continue;
    mai_resolver_add_package (resolver, package);
    clutter_model_append (priv->store, 0, package, -1);
    g_object_unref (package);
  }
  mai_parser_iterator_destroy (iter);

  /* Match app list to package database. */
  g_signal_connect (resolver, "notify::progress",
                    G_CALLBACK (_resolver_notify_progress_cb), self);
  mai_resolver_resolve (resolver, &error);
  if (error)
  {
    /* FIXME propagate to GUI. */
    g_warning ("%s : %s", G_STRLOC, error->message);
    g_clear_error (&error);
  }

  /* Build category list. */
  priv->categories = NULL;
  for (iter = mai_parser_get_category_iterator (parser);
       !mai_parser_iterator_is_end (iter);
       mai_parser_iterator_next (iter))
  {
    MaiCategory *category = mai_parser_iterator_create_category (iter);
    if (!category)
      continue;
    /* List takes reference. */
    priv->categories = g_list_append (priv->categories, category);
  }
  mai_parser_iterator_destroy (iter);

  g_object_unref (parser);
  return G_OBJECT (self);
}

static void
mai_model_class_init (MaiModelClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GParamFlags param_flags;

  g_type_class_add_private (klass, sizeof (MaiModelPrivate));

  object_class->constructor = _constructor;
  object_class->get_property = _get_property;
  object_class->set_property = _set_property;
  object_class->dispose = _dispose;

  param_flags = G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_NAME;

  g_object_class_install_property (object_class,
                                   PROP_APPSTORE_FILE,
                                   g_param_spec_string ("appstore-file",
                                                        "Appstore file",
                                                        "File to load package info from",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_STORE,
                                   g_param_spec_object ("store",
                                                        "Store",
                                                        "Appstore model",
                                                        MAI_TYPE_OBJECT_STORE,
                                                        param_flags));
  g_object_class_install_property (object_class,
                                   PROP_CATEGORIES,
                                   g_param_spec_pointer ("categories",
                                                        "Categories",
                                                        "Appstore cateogories",
                                                        param_flags));
}

static void
mai_model_init (MaiModel *self)
{
}

MaiModel *
mai_model_new (const gchar *appstore_file)
{
  MaiModel *self;

  g_return_val_if_fail (appstore_file, NULL);

  self = g_object_new (MAI_TYPE_MODEL,
                       "appstore-file", appstore_file,
                       NULL);

  return self;
}

const gchar *
mai_model_get_appstore_file (MaiModel const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->appstore_file;
}

ClutterModel *
mai_model_get_store (MaiModel const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->store;
}

GList const *
mai_model_get_categories (MaiModel const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->categories;
}

/*
 * Doesn't work consistently when a filter is set.
 */
GList *
mai_model_get_used_categories (MaiModel const *self)
{
  MaiModelPrivate *priv;
  GList           *used_categories = NULL;

  g_return_val_if_fail (self, NULL);

  priv = GET_PRIVATE (self);

  /* Iterate categories to check which have packages. */
  for (GList const *list = priv->categories;
       list;
       list = list->next)
  {
    MaiCategory *category = MAI_CATEGORY (list->data);
    const gchar *category_name = mai_category_get_name (category);
    ClutterModelIter *iter = NULL;

    /* Iterate model to find package belonging to the current category. */
    for (iter = clutter_model_get_first_iter (priv->store);
         iter && !clutter_model_iter_is_last (iter);
         iter = clutter_model_iter_next (iter))
    {
      MaiPackage *package = NULL;
      clutter_model_iter_get (iter, 0, &package, -1);

      /* Iterate package's categories to check whether it's in the current one. */
      for (const gchar **categories = mai_package_get_categories (package);
           *categories;
           categories++)
      {
        if (0 == g_strcmp0 (category_name, *categories))
        {
          used_categories = g_list_append (used_categories, category);
          goto next_category;
        }
      }
    }
next_category:
    if (iter)
      g_object_unref (iter);
  }

  return used_categories;
}

static gboolean
_filter_category (ClutterModel      *model,
                  ClutterModelIter  *iter,
                  MaiCategory const *category)
{
  const gchar *category_name = mai_category_get_name (category);
  MaiPackage  *package = NULL;
  gboolean     matched = FALSE;

  clutter_model_iter_get (iter, 0, &package, -1);
  g_warn_if_fail (package);
  if (package)
  {
    for (const gchar **categories = mai_package_get_categories (package);
         categories && *categories;
         categories++)
    {
      matched = 0 == g_strcmp0 (category_name, *categories);
      if (matched)
        break;
    }
  }

  return matched;
}

void
mai_model_filter_category (MaiModel           *self,
                           MaiCategory const  *category)
{
  MaiModelPrivate *priv;

  g_return_if_fail (self);

  priv = GET_PRIVATE (self);

  if (category)
  {
    clutter_model_set_filter (priv->store,
                              (ClutterModelFilterFunc) _filter_category,
                              (gpointer) category,
                              NULL);
  } else {
    mai_model_filter_clear (self);
  }
}

static gboolean
_filter_text (ClutterModel      *model,
              ClutterModelIter  *iter,
              const gchar       *needle)
{
  MaiPackage *package = NULL;

  /* Clutter seems to need this to not assert. */
  if (clutter_model_iter_is_last (iter))
    return TRUE;

  /* This is rather inefficient, cache down-cased strings if too slow. */
  clutter_model_iter_get (iter, 0, &package, -1);
  if (package)
  {
    gchar     *haystack;
    gboolean   ret;

    /* Match package name. */
    haystack = g_utf8_strdown (mai_package_get_name (package), -1);
    ret = (gboolean) g_strstr_len (haystack, -1, needle);
    g_free (haystack);

    if (ret)
      return TRUE;

    /* Match package description. */
    haystack = g_utf8_strdown (mai_package_get_description (package), -1);
    ret = (gboolean) g_strstr_len (haystack, -1, needle);
    g_free (haystack);

    if (ret)
      return TRUE;
  }

  return FALSE;
}

void
mai_model_filter_text (MaiModel     *self,
                       const gchar  *text)
{
  MaiModelPrivate *priv;

  g_return_if_fail (self);

  priv = GET_PRIVATE (self);

  if (text)
  {
    clutter_model_set_filter (priv->store,
                              (ClutterModelFilterFunc) _filter_text,
                              g_utf8_strdown (text, -1),
                              (GDestroyNotify) g_free);
  } else {
    mai_model_filter_clear (self);
  }
}

void
mai_model_filter_clear (MaiModel *self)
{
  MaiModelPrivate *priv;

  g_return_if_fail (self);

  priv = GET_PRIVATE (self);
  clutter_model_set_filter (priv->store, NULL, NULL, NULL);
}

MaiPackage *
mai_model_lookup_package (MaiModel    *self,
                          const gchar *package_name)
{
  MaiModelPrivate   *priv = GET_PRIVATE (self);
  ClutterModelIter  *iter;
  MaiPackage        *package;

  g_return_val_if_fail (package_name, NULL);

  for (iter = clutter_model_get_first_iter (priv->store);
       !clutter_model_iter_is_last (iter);
       iter = clutter_model_iter_next (iter))
  {
    package = NULL;
    clutter_model_iter_get (iter, 0, &package, -1);
    if (package &&
        0 == g_strcmp0 (mai_package_get_package_name (package), package_name))
    {
      break;
    } else {
      package = NULL;
    }
  }

  g_object_unref (iter);
  return package;
}

