
/*
 * moblin-app-installer -- Moblin garage client application.
 * Copyright (c) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Author: Robert Staudinger <robertx.staudinger@intel.com>
 * Based on clutter-list-model.c
 *    Copyright (C) 2006 OpenedHand, by
 *    Matthew Allum  <mallum@openedhand.com>
 *    Neil Jagdish Patel <njp@o-hand.com>
 *    Emmanuele Bassi <ebassi@openedhand.com>
 */

#include "mai-object-store.h"
#include "mai-utils.h"
#include "config.h"

typedef struct MaiObjectStoreIter_ MaiObjectStoreIter;

static void
mai_object_store_object_property_notify (GObject         *object,
                                         GParamSpec      *pspec,
                                         MaiObjectStore  *self);

static void
mai_object_store_detach_object          (MaiObjectStore  *self,
                                         GObject         *object);

static void
mai_object_store_attach_object          (MaiObjectStore  *self,
                                         GObject         *object);

/*
 * MaiObjectStore declaration.
 */

G_DEFINE_TYPE (MaiObjectStore, mai_object_store, CLUTTER_TYPE_MODEL)

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), MAI_TYPE_OBJECT_STORE, MaiObjectStorePrivate))

typedef struct
{
  GSequence           *sequence;
  MaiObjectStoreIter  *cached_iter;
} MaiObjectStorePrivate;

/*
 * MaiObjectStoreIter.
 */

#define MAI_TYPE_OBJECT_STORE_ITER mai_object_store_iter_get_type()

#define MAI_OBJECT_STORE_ITER(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST ((obj), \
  MAI_TYPE_OBJECT_STORE_ITER, MaiObjectStoreIter))

#define MAI_OBJECT_STORE_ITER_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST ((klass), \
  MAI_TYPE_OBJECT_STORE_ITER, MaiObjectStoreIterClass))

#define MAI_IS_OBJECT_STORE_ITER(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE ((obj), \
  MAI_TYPE_OBJECT_STORE_ITER))

#define MAI_IS_OBJECT_STORE_ITER_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE ((klass), \
  MAI_TYPE_OBJECT_STORE_ITER))

#define MAI_OBJECT_STORE_ITER_GET_CLASS(obj) \
  (G_TYPE_INSTANCE_GET_CLASS ((obj), \
  MAI_TYPE_OBJECT_STORE_ITER, MaiObjectStoreIterClass))

struct MaiObjectStoreIter_ {
  ClutterModelIter   parent;
  GSequenceIter     *seq_iter;  /* NULL means it's the end iter. */
};

typedef struct {
  ClutterModelIterClass parent;
} MaiObjectStoreIterClass;

GType mai_object_store_iter_get_type (void) G_GNUC_CONST;

G_DEFINE_TYPE (MaiObjectStoreIter, mai_object_store_iter, CLUTTER_TYPE_MODEL_ITER)

static guint
mai_object_store_iter_get_row (ClutterModelIter *iter)
{
  MaiObjectStoreIter    *self;
  ClutterModel          *store;
  MaiObjectStorePrivate *priv;
  MaiObjectStoreIter    *cached_iter;
  gint                   row = -1;

  g_return_val_if_fail (MAI_IS_OBJECT_STORE_ITER (iter), FALSE);

  self = MAI_OBJECT_STORE_ITER (iter);
  store = clutter_model_iter_get_model (iter);
  priv = GET_PRIVATE (store);
  cached_iter = priv->cached_iter;

  if (clutter_model_get_filter_set (store))
    {
      /* Count matching rows using the cached_iter until hitting self. */
      gboolean found = FALSE;
      for (cached_iter->seq_iter = g_sequence_get_begin_iter (priv->sequence);
           !g_sequence_iter_is_end (cached_iter->seq_iter);
           cached_iter->seq_iter = g_sequence_iter_next (cached_iter->seq_iter))
        {
          if (clutter_model_filter_iter (store, CLUTTER_MODEL_ITER (cached_iter)))
            {
              row++;
            }
          if (self->seq_iter &&
              0 == g_sequence_iter_compare (self->seq_iter, cached_iter->seq_iter))
            {
              found = TRUE;
              break;
            }
        }
        if (!found)
          row = -1;
    }
  else
    {
      row = g_sequence_iter_get_position (self->seq_iter);
    }

  /* Silently accept the unmatched iters. */
  return row >= 0 ? row : 0;
}

static void
mai_object_store_iter_get_value (ClutterModelIter *iter,
                                 guint             column,
                                 GValue           *value)
{
  MaiObjectStoreIter  *self;
  GObject             *object;
  const gchar         *name;

  g_return_if_fail (MAI_IS_OBJECT_STORE_ITER (iter));
  g_return_if_fail (value);

  self = MAI_OBJECT_STORE_ITER (iter);
  if (!self->seq_iter)
    return;
  object = g_sequence_get (self->seq_iter);

  if (column == 0)
    {
      g_value_set_object (value, object);
    }
  else
    {
      name = clutter_model_get_column_name (clutter_model_iter_get_model (iter),
                                            column);
      g_object_get_property (object, name, value);
    }
}

static void
mai_object_store_iter_set_value (ClutterModelIter *iter,
                                 guint             column,
                                 const GValue     *value)
{
  MaiObjectStoreIter  *self;
  ClutterModel        *model;
  GObject             *object;

  g_return_if_fail (MAI_IS_OBJECT_STORE_ITER (iter));
  g_return_if_fail (value);

  self = MAI_OBJECT_STORE_ITER (iter);
  model = clutter_model_iter_get_model (iter);
  object = g_sequence_get (self->seq_iter);

  if (column == 0)
    {
      /* Set "master" column. NULL is legal since append() does an empty row. */
      if (object)
        {
          mai_object_store_detach_object (MAI_OBJECT_STORE (model), object);
          g_object_unref (object);
        }

      object = g_value_get_object (value);
      g_sequence_set (self->seq_iter, g_object_ref (object));

      /* Hook up "changed" notifications for the object's properties. */
      mai_object_store_attach_object (MAI_OBJECT_STORE (model), object);
    }
  else if (object)
    {
      const gchar *name = clutter_model_get_column_name (model, column);
      g_object_set_property (object, name, value);
    }
  else
    {
      g_warning ("%s Cannot set column %d on NULL object",
                 G_STRLOC,
                 column);
    }
}

static gboolean
mai_object_store_iter_is_first (ClutterModelIter *iter)
{
  MaiObjectStoreIter    *self;
  ClutterModel          *store;
  MaiObjectStorePrivate *priv;
  MaiObjectStoreIter    *cached_iter;
  gboolean               is_seq_first;
  gboolean               is_first = TRUE;

  g_return_val_if_fail (MAI_IS_OBJECT_STORE_ITER (iter), FALSE);

  self = MAI_OBJECT_STORE_ITER (iter);
  store = clutter_model_iter_get_model (iter);
  priv = GET_PRIVATE (store);
  cached_iter = priv->cached_iter;

  /* Go backwards from iter looking for non-filtered rows. */
  for (is_seq_first = g_sequence_iter_is_begin (self->seq_iter),
       cached_iter->seq_iter = g_sequence_iter_prev (self->seq_iter);
       !is_seq_first;
       is_seq_first = g_sequence_iter_is_begin (cached_iter->seq_iter),
       cached_iter->seq_iter = g_sequence_iter_prev (cached_iter->seq_iter))
    {
      if (clutter_model_filter_iter (store, CLUTTER_MODEL_ITER (cached_iter)))
        {
          is_first = FALSE;
          break;
        }
    }

  return is_first;
}

static gboolean
mai_object_store_iter_is_last (ClutterModelIter *iter)
{
  g_return_val_if_fail (MAI_IS_OBJECT_STORE_ITER (iter), TRUE);

  /* FIXME: maybe also check for end-of-sequence. */
  return MAI_OBJECT_STORE_ITER (iter)->seq_iter == NULL;
}

static ClutterModelIter *
mai_object_store_iter_next (ClutterModelIter *iter)
{
  MaiObjectStoreIter    *self;
  ClutterModel          *store;
  MaiObjectStorePrivate *priv;
  MaiObjectStoreIter    *cached_iter;

  g_return_val_if_fail (MAI_IS_OBJECT_STORE_ITER (iter), NULL);

  self = MAI_OBJECT_STORE_ITER (iter);
  store = clutter_model_iter_get_model (iter);
  priv = GET_PRIVATE (store);
  cached_iter = priv->cached_iter;

  /* Look for next non-filtered row. */
  cached_iter->seq_iter = self->seq_iter;
  self->seq_iter = NULL;
  for (cached_iter->seq_iter = g_sequence_iter_next (cached_iter->seq_iter);
       !g_sequence_iter_is_end (cached_iter->seq_iter);
       cached_iter->seq_iter = g_sequence_iter_next (cached_iter->seq_iter))
    {
      if (clutter_model_filter_iter (store, CLUTTER_MODEL_ITER (cached_iter)))
        {
          self->seq_iter = cached_iter->seq_iter;
          break;
        }
    }

  return iter;
}

static ClutterModelIter *
mai_object_store_iter_prev (ClutterModelIter *iter)
{
  MaiObjectStoreIter    *self;
  ClutterModel          *store;
  MaiObjectStorePrivate *priv;
  MaiObjectStoreIter    *cached_iter;
  GSequenceIter         *initial_seq_iter;
  gboolean               at_begin;

  g_return_val_if_fail (MAI_IS_OBJECT_STORE_ITER (iter), NULL);

  self = MAI_OBJECT_STORE_ITER (iter);
  store = clutter_model_iter_get_model (iter);
  priv = GET_PRIVATE (store);
  cached_iter = priv->cached_iter;
  initial_seq_iter = self->seq_iter;

  /* Look for prev non-filtered row. */
  cached_iter->seq_iter = self->seq_iter;
  self->seq_iter = NULL;
  for (at_begin = g_sequence_iter_is_begin (cached_iter->seq_iter),
       cached_iter->seq_iter = g_sequence_iter_prev (cached_iter->seq_iter);
       !at_begin;
       at_begin = g_sequence_iter_is_begin (cached_iter->seq_iter),
       cached_iter->seq_iter =  g_sequence_iter_prev (cached_iter->seq_iter))
    {
      if (clutter_model_filter_iter (store, CLUTTER_MODEL_ITER (cached_iter)))
        {
          self->seq_iter = cached_iter->seq_iter;
          break;
        }
    }

  /* Return same iter if not found. */
  if (self->seq_iter == NULL)
    self->seq_iter = initial_seq_iter;

  return iter;
}

static ClutterModelIter *
mai_object_store_iter_copy (ClutterModelIter *iter)
{
  MaiObjectStoreIter  *self;
  ClutterModelIter    *copy;

  g_return_val_if_fail (MAI_IS_OBJECT_STORE_ITER (iter), NULL);

  self = MAI_OBJECT_STORE_ITER (iter);
  copy = g_object_new (MAI_TYPE_OBJECT_STORE_ITER,
                       "model", clutter_model_iter_get_model (iter),
                       NULL);

  /* this is safe, because the seq_iter pointer on the passed
   * iterator will be always be overwritten in ::next or ::prev */
  MAI_OBJECT_STORE_ITER (copy)->seq_iter = self->seq_iter;

  return copy;
}

static void
mai_object_store_iter_class_init (MaiObjectStoreIterClass *klass)
{
  ClutterModelIterClass *iter_class = CLUTTER_MODEL_ITER_CLASS (klass);

  iter_class->get_row   = mai_object_store_iter_get_row;
  iter_class->get_value = mai_object_store_iter_get_value;
  iter_class->set_value = mai_object_store_iter_set_value;
  iter_class->is_first  = mai_object_store_iter_is_first;
  iter_class->is_last   = mai_object_store_iter_is_last;
  iter_class->next      = mai_object_store_iter_next;
  iter_class->prev      = mai_object_store_iter_prev;
  iter_class->copy      = mai_object_store_iter_copy;
}

static void
mai_object_store_iter_init (MaiObjectStoreIter *iter)
{
  iter->seq_iter = NULL;
}

/*
 * MaiObjectStore.
 */

static void
mai_object_store_finalize (GObject *gobject)
{
  MaiObjectStorePrivate *priv = GET_PRIVATE (gobject);

  g_sequence_free (priv->sequence);
  priv->sequence = NULL;

  G_OBJECT_CLASS (mai_object_store_parent_class)->finalize (gobject);
}

static void
_detach_if_non_null (GObject        *object,
                     MaiObjectStore *self)
{
  if (G_IS_OBJECT (object))
    mai_object_store_detach_object (self, object);
}

static void
mai_object_store_dispose (GObject *gobject)
{
  MaiObjectStorePrivate *priv = GET_PRIVATE (gobject);

  g_sequence_foreach_range (g_sequence_get_begin_iter (priv->sequence),
                            g_sequence_get_end_iter (priv->sequence),
                            (GFunc) _detach_if_non_null,
                            gobject);

  g_sequence_remove_range (g_sequence_get_begin_iter (priv->sequence),
                           g_sequence_get_end_iter (priv->sequence));

  if (priv->cached_iter)
    {
      g_object_unref (priv->cached_iter);
      priv->cached_iter = NULL;
    }

  G_OBJECT_CLASS (mai_object_store_parent_class)->dispose (gobject);
}

static ClutterModelIter *
mai_object_store_get_iter_at_row (ClutterModel *self,
                                  guint         row)
{
  MaiObjectStorePrivate *priv;
  MaiObjectStoreIter    *cached_iter;
  MaiObjectStoreIter    *iter = NULL;
  gint                   r = -1;

  g_return_val_if_fail (MAI_IS_OBJECT_STORE (self), NULL);

  priv = GET_PRIVATE (self);
  cached_iter = priv->cached_iter;

  /* Work the cached iter to defer object instantiation until success. */
  if (clutter_model_get_filter_set (self))
    {
      /* Count matching rows. */
      for (cached_iter->seq_iter = g_sequence_get_begin_iter (priv->sequence);
           !g_sequence_iter_is_end (cached_iter->seq_iter);
           cached_iter->seq_iter = g_sequence_iter_next (cached_iter->seq_iter))
        {
          if (clutter_model_filter_iter (self, CLUTTER_MODEL_ITER (cached_iter)))
            {
              r++;
              if ((unsigned) r == row)
                break;
            }
        }
    }
  else
    {
      r = row;
      cached_iter->seq_iter = g_sequence_get_iter_at_pos (priv->sequence, row);
    }

  if (r > -1)
    {
      iter = g_object_new (MAI_TYPE_OBJECT_STORE_ITER,
                            "model", self,
                            NULL);
      iter->seq_iter = cached_iter->seq_iter;
    }

  return (ClutterModelIter *) iter;
}

static ClutterModelIter *
mai_object_store_insert_row (ClutterModel *self,
                             gint          index_)
{
  MaiObjectStorePrivate *priv;
  MaiObjectStoreIter    *iter;

  g_return_val_if_fail (MAI_IS_OBJECT_STORE (self), NULL);

  priv = GET_PRIVATE (self);
  iter = g_object_new (MAI_TYPE_OBJECT_STORE_ITER,
                       "model", self,
                       NULL);

  if (index_ < 0)
    {
      iter->seq_iter = g_sequence_append (priv->sequence, NULL);
    }
  else
    {
      GSequenceIter *seq_iter;
      seq_iter = g_sequence_get_iter_at_pos (priv->sequence, index_);
      iter->seq_iter = g_sequence_insert_before (seq_iter, NULL);
    }

  return CLUTTER_MODEL_ITER (iter);
}

static void
mai_object_store_remove_row (ClutterModel *self,
                             guint         row)
{
  MaiObjectStorePrivate *priv;
  MaiObjectStoreIter    *iter;

  g_return_if_fail (MAI_IS_OBJECT_STORE (self));

  priv = GET_PRIVATE (self);
  iter = g_object_new (MAI_TYPE_OBJECT_STORE_ITER,
                       "model", self,
                       NULL);
  iter->seq_iter = g_sequence_get_iter_at_pos (priv->sequence, row);

  /* the actual row is removed from the sequence inside
   * the ::row-removed signal class handler, so that every
   * handler connected to ::row-removed will still get
   * a valid iterator, and every signal connected to
   * ::row-removed with the AFTER flag will get an updated
   * store */
  g_signal_emit_by_name (self, "row-removed", iter);
  g_object_unref (iter);
}

typedef struct
{
  ClutterModel          *store;
  const gchar           *property_name;
  ClutterModelSortFunc   func;
  gpointer               data;
} SortClosure;

static gint
sort_store_default (gconstpointer a,
                    gconstpointer b,
                    gpointer      data)
{
  SortClosure *closure = data;
  GValue       p1 = { 0, };
  GValue       p2 = { 0, };

  if (a == NULL && b == NULL)
    return 0;
  else if (a == NULL)
    return -1;
  else if (b == NULL)
    return 1;

  g_object_get_property (G_OBJECT (a), closure->property_name, &p1);
  g_object_get_property (G_OBJECT (b), closure->property_name, &p2);

  return closure->func (closure->store, &p1, &p2, closure->data);
}

static void
mai_object_store_resort (ClutterModel         *self,
                         ClutterModelSortFunc  func,
                         gpointer              data)
{
  MaiObjectStorePrivate *priv;
  SortClosure            closure;
  gint                   column;

  g_return_if_fail (MAI_IS_OBJECT_STORE (self));

  priv = GET_PRIVATE (self);
  column = clutter_model_get_sorting_column (self);

  closure.store = self;
  closure.property_name = clutter_model_get_column_name (self, column);
  closure.func = func;
  closure.data = data;

  g_sequence_sort (priv->sequence, sort_store_default, &closure);
}

static void
mai_object_store_row_removed (ClutterModel     *self,
                              ClutterModelIter *iter_)
{
  MaiObjectStoreIter  *iter;
  GSequenceIter       *seq_iter;
  GObject             *object;

  iter = MAI_OBJECT_STORE_ITER (iter_);
  seq_iter = iter->seq_iter;
  iter->seq_iter = g_sequence_iter_next (iter->seq_iter);
  object = g_sequence_get (iter->seq_iter);

  if (G_IS_OBJECT (object))
    {
      mai_object_store_detach_object (MAI_OBJECT_STORE (self), object);
      g_object_unref (object);
    }

  g_sequence_remove (seq_iter);
}

static void
mai_object_store_class_init (MaiObjectStoreClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  ClutterModelClass *store_class = CLUTTER_MODEL_CLASS (klass);

  g_type_class_add_private (klass, sizeof (MaiObjectStorePrivate));

  gobject_class->dispose = mai_object_store_dispose;
  gobject_class->finalize = mai_object_store_finalize;

  store_class->get_iter_at_row = mai_object_store_get_iter_at_row;
  store_class->insert_row      = mai_object_store_insert_row;
  store_class->remove_row      = mai_object_store_remove_row;
  store_class->resort          = mai_object_store_resort;

  store_class->row_removed     = mai_object_store_row_removed;
}

/*
 * The destroy function is called regardless of NULL data.
 */
static void
_unref_if_non_null (GObject *object)
{
  if (G_IS_OBJECT (object))
    g_object_unref (object);
}

static void
mai_object_store_init (MaiObjectStore *self)
{
  MaiObjectStorePrivate *priv = GET_PRIVATE (self);

  priv->sequence = g_sequence_new ((GDestroyNotify) _unref_if_non_null);
  priv->cached_iter = g_object_new (MAI_TYPE_OBJECT_STORE_ITER,
                                    "model", self,
                                    NULL);
}

ClutterModel *
mai_object_store_new (guint n_columns,
                      ...)
{
  ClutterModel  *self;
  GType          types[n_columns];
  const gchar   *names[n_columns];
  va_list        args;

  g_return_val_if_fail (n_columns > 0, NULL);

  self = g_object_new (MAI_TYPE_OBJECT_STORE, NULL);

  va_start (args, n_columns);

  for (guint i = 0; i < n_columns; i++)
    {
      types[i] = va_arg (args, GType);
      names[i] = va_arg (args, gchar *);
    }

  va_end (args);

  clutter_model_set_types (self, n_columns, types);
  clutter_model_set_names (self, n_columns, names);

  return self;
}

static void
mai_object_store_object_property_notify (GObject         *object,
                                         GParamSpec      *pspec,
                                         MaiObjectStore  *self)
{
  MaiObjectStorePrivate *priv = GET_PRIVATE (self);
  GSequenceIter         *notify_seq_iter = NULL;
  MaiObjectStoreIter    *iter;

  /* Find corresponding row of changed object. */
  for (GSequenceIter *seq_iter = g_sequence_get_begin_iter (priv->sequence);
       !g_sequence_iter_is_end (seq_iter);
       seq_iter = g_sequence_iter_next (seq_iter))
    {
      if (object == g_sequence_get (seq_iter))
        {
          notify_seq_iter = seq_iter;
          break;
        }
    }

  g_return_if_fail (notify_seq_iter);

  iter = g_object_new (MAI_TYPE_OBJECT_STORE_ITER,
                       "model", self,
                       NULL);
  iter->seq_iter = notify_seq_iter;
  g_signal_emit_by_name (self, "row-changed", iter);
  g_object_unref (iter);
}

static void
mai_object_store_attach_object (MaiObjectStore  *self,
                                GObject         *object)
{
  guint        n_columns;
  const gchar *column_name;
  gchar       *signal_name;

  /* Start at column 1 because 0 hold the actual object. */
  n_columns = clutter_model_get_n_columns (CLUTTER_MODEL (self));
  for (guint i = 1; i < n_columns; i++)
    {
      column_name = clutter_model_get_column_name (CLUTTER_MODEL (self), i);
      signal_name = g_strdup_printf ("notify::%s", column_name);
      g_signal_connect (object,
                        signal_name,
                        G_CALLBACK (mai_object_store_object_property_notify),
                        self);
      g_free (signal_name);
    }
}

static void
mai_object_store_detach_object (MaiObjectStore  *self,
                                GObject         *object)
{
  g_signal_handlers_disconnect_by_func (object,
      MAI_CALLBACK_TO_POINTER (mai_object_store_object_property_notify),
      self);
}

void
mai_object_store_foreach_unfiltered (MaiObjectStore          *self,
                                     ClutterModelForeachFunc  func,
                                     gpointer                 user_data)
{
  MaiObjectStorePrivate *priv;

  g_return_if_fail (MAI_IS_OBJECT_STORE (self));
  g_return_if_fail (func);

  priv = GET_PRIVATE (self);

  for (GSequenceIter *seq_iter = g_sequence_get_begin_iter (priv->sequence);
       !g_sequence_iter_is_end (seq_iter);
       seq_iter = g_sequence_iter_next (seq_iter))
    {
      priv->cached_iter->seq_iter = seq_iter;
      if (!func ((ClutterModel *) self,
                 (ClutterModelIter *) priv->cached_iter,
                 user_data))
        {
          break;
        }
    }
}

