
/*
 * moblin-app-installer -- Moblin garage client application.
 * Copyright (c) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Author: Robert Staudinger <robertx.staudinger@intel.com>
 * Based on code by Arjan van de Ven <arjan@linux.intel.com>
 */

#include <string.h>
#include <glib/gi18n.h>
#include <libxml/tree.h>
#include "mai-category.h"
#include "mai-package.h"
#include "mai-parser.h"
#include "config.h"

G_DEFINE_TYPE (MaiParser, mai_parser, G_TYPE_OBJECT)

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), MAI_TYPE_PARSER, MaiParserPrivate))

enum
{
  PROP_0,

  PROP_APPSTORE_FILE
};

typedef struct
{
  /* Properties */
  gchar       *appstore_file;
  /* Data */
  xmlDocPtr    appstore_doc;
} MaiParserPrivate;

struct MaiParserIterator_
{
  gchar       *tag_name;
  xmlNodePtr   node;
};

static void
_get_property (GObject    *object,
               guint       property_id,
               GValue     *value,
               GParamSpec *pspec)
{
  switch (property_id) {
  case PROP_APPSTORE_FILE:
    g_value_set_string (value,
                        mai_parser_get_appstore_file (MAI_PARSER (object)));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_set_property (GObject      *object,
               guint         property_id,
               const GValue *value,
               GParamSpec   *pspec)
{
  MaiParserPrivate *priv = GET_PRIVATE (object);

  switch (property_id) {
  case PROP_APPSTORE_FILE:
    /* Construct-only, no need to notify and stuff. */
    priv->appstore_file = g_value_dup_string (value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_dispose (GObject *object)
{
  MaiParserPrivate *priv = GET_PRIVATE (object);

  if (priv->appstore_file)
    g_free (priv->appstore_file), priv->appstore_file = NULL;

  if (priv->appstore_doc)
    xmlFreeDoc (priv->appstore_doc), priv->appstore_doc = NULL;

  G_OBJECT_CLASS (mai_parser_parent_class)->dispose (object);
}

static GObject *
_constructor (GType                  type,
              guint                  n_properties,
              GObjectConstructParam *properties)
{
  MaiParser *self = (MaiParser *)
                      G_OBJECT_CLASS (mai_parser_parent_class)
                          ->constructor (type, n_properties, properties);
  MaiParserPrivate *priv = GET_PRIVATE (self);
  xmlNodePtr        node;

  if (priv->appstore_file == NULL)
  {
    g_critical ("%s : %s", G_STRLOC, "No appstore file");
    goto bail;
  }

  priv->appstore_doc = xmlParseFile (priv->appstore_file);
  if (!priv->appstore_doc)
  {
    g_critical ("Failed to parse '%s'", priv->appstore_file);
    goto bail;
  }

  node = xmlDocGetRootElement (priv->appstore_doc);
  if (!node)
  {
    g_critical ("Empty appstore file '%s'", priv->appstore_file);
    goto bail;
  }

  if (0 != g_strcmp0 (node->name, "applist"))
  {
    g_critical ("Appstore file invalid, root node '%s'", node->name);
    goto bail;
  }

  return G_OBJECT (self);

bail:
  if (priv->appstore_doc)
    xmlFreeDoc (priv->appstore_doc), priv->appstore_doc = NULL;
  return NULL; /* Is this legal? Unref self? */
}

static void
mai_parser_class_init (MaiParserClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GParamFlags param_flags;

  g_type_class_add_private (klass, sizeof (MaiParserPrivate));

  object_class->constructor = _constructor;
  object_class->get_property = _get_property;
  object_class->set_property = _set_property;
  object_class->dispose = _dispose;

  param_flags = G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_NAME;

  g_object_class_install_property (object_class,
                                   PROP_APPSTORE_FILE,
                                   g_param_spec_string ("appstore-file",
                                                        "Appstore file",
                                                        "File to load package info from",
                                                        NULL, param_flags));
}

static void
mai_parser_init (MaiParser *self)
{
}

MaiParser *
mai_parser_new (const gchar *appstore_file)
{
  MaiParser *self;

  g_return_val_if_fail (appstore_file, NULL);

  self = g_object_new (MAI_TYPE_PARSER,
                       "appstore-file", appstore_file,
                       NULL);

  return self;
}

const gchar *
mai_parser_get_appstore_file (MaiParser const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->appstore_file;
}

MaiParserIterator *
mai_parser_get_category_iterator (MaiParser const *self)
{
  MaiParserPrivate  *priv;
  MaiParserIterator *iter = NULL;
  xmlNodePtr         node;

  g_return_val_if_fail (self, NULL);

  priv = GET_PRIVATE (self);
  g_return_val_if_fail (priv->appstore_doc, NULL);

  node = xmlDocGetRootElement (priv->appstore_doc);
  if (!node)
  {
    g_critical ("Empty appstore file '%s'", priv->appstore_file);
    return NULL;
  }

  if (0 != g_strcmp0 (node->name, "applist"))
  {
    g_critical ("Appstore file invalid, root node '%s'", node->name);
    return NULL;
  }

  node = node->xmlChildrenNode;
  if (!node)
  {
    g_critical ("Appstore is empty");
    return NULL;
  }

  while (node)
  {
    if (0 == g_strcmp0 (node->name, "category"))
    {
      iter = g_new0 (MaiParserIterator, 1);
      iter->tag_name = g_strdup ("category");
      iter->node = node;
      break;
    }
    node = node->next;
  }

  if (!iter)
    g_critical ("Appstore file invalid, no 'category' node found");

  return iter;
}

MaiParserIterator *
mai_parser_get_package_iterator (MaiParser const *self)
{
  MaiParserPrivate  *priv;
  MaiParserIterator *iter = NULL;
  xmlNodePtr         node;

  g_return_val_if_fail (self, NULL);

  priv = GET_PRIVATE (self);
  g_return_val_if_fail (priv->appstore_doc, NULL);

  node = xmlDocGetRootElement (priv->appstore_doc);
  if (!node)
  {
    g_critical ("Empty appstore file '%s'", priv->appstore_file);
    return NULL;
  }

  if (0 != g_strcmp0 (node->name, "applist"))
  {
    g_critical ("Appstore file invalid, root node '%s'", node->name);
    return NULL;
  }

  node = node->xmlChildrenNode;
  if (!node)
  {
    g_critical ("Appstore is empty");
    return NULL;
  }

  while (node)
  {
    if (0 == g_strcmp0 (node->name, "application"))
    {
      iter = g_new0 (MaiParserIterator, 1);
      iter->tag_name = g_strdup ("application");
      iter->node = node;
      break;
    }
    node = node->next;
  }

  if (!iter)
    g_critical ("Appstore file invalid, no 'application' node found");

  return iter;
}

/*
 * Iterator
 */

gboolean
mai_parser_iterator_next (MaiParserIterator *iter)
{
  g_return_val_if_fail (iter, FALSE);
  g_return_val_if_fail (iter->node, FALSE);

  /* Look for next tag that matches this iter's type. */
  for (iter->node = iter->node->next;
       iter->node && 0 != g_strcmp0 (iter->tag_name, iter->node->name);
       iter->node = iter->node->next)
    ;

  return (gboolean) iter->node;
}

MaiCategory *
mai_parser_iterator_create_category (MaiParserIterator const *iter)
{
  MaiCategory *category;
  gboolean     is_default;

  g_return_val_if_fail (iter, NULL);
  g_return_val_if_fail (iter->node, NULL);
  g_return_val_if_fail (0 == g_strcmp0 (iter->node->name, "category"), NULL);

  is_default = 0 == g_strcmp0 (xmlGetProp (iter->node, "default"), "true") ?
                  TRUE :
                  FALSE;

  category = mai_category_new (xmlGetProp (iter->node, "name"),
                               xmlGetProp (iter->node, "icon"),
                               is_default);

  return category;
}

MaiPackage *
mai_parser_iterator_create_package (MaiParserIterator const *iter)
{
  MaiPackage   *package;
  gchar       **categories;
  const gchar  *package_name;

  g_return_val_if_fail (iter, NULL);
  g_return_val_if_fail (iter->node, NULL);
  g_return_val_if_fail (0 == g_strcmp0 (iter->node->name, "application"), NULL);

  package_name = xmlGetProp (iter->node, "package");
  g_return_val_if_fail (package_name && strlen (package_name) > 0, NULL);

  categories = g_strsplit (xmlGetProp (iter->node, "categories"), ";", -1);

  package = mai_package_new (package_name,
                             xmlGetProp (iter->node, "name"),
                             xmlGetProp (iter->node, "version"),
                             xmlGetProp (iter->node, "description"),
                             xmlGetProp (iter->node, "long_description"),
                             xmlGetProp (iter->node, "icon"),
                             (const gchar **) categories);

  g_strfreev (categories);

  return package;
}

gboolean
mai_parser_iterator_is_end (MaiParserIterator const *iter)
{
  g_return_val_if_fail (iter, TRUE);

  return iter->node == NULL;
}

void
mai_parser_iterator_destroy (MaiParserIterator *iter)
{
  g_return_if_fail (iter);

  g_free (iter->tag_name);
  g_free (iter);
}

