
/*
 * moblin-app-installer -- Moblin garage client application.
 * Copyright (c) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Author: Robert Staudinger <robertx.staudinger@intel.com>
 */

#include "mai-transaction.h"
#include "mai-transaction-manager.h"
#include "mai-utils.h"

G_DEFINE_TYPE (MaiTransactionManager, mai_transaction_manager, G_TYPE_OBJECT)

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), MAI_TYPE_TRANSACTION_MANAGER, MaiTransactionManagerPrivate))

typedef struct
{
  GQueue *queue;
} MaiTransactionManagerPrivate;

static void
_dispose (GObject *object)
{
  MaiTransactionManagerPrivate *priv = GET_PRIVATE (object);

  if (priv->queue)
  {
    g_queue_foreach (priv->queue, (GFunc) g_object_unref, NULL);
    g_queue_free (priv->queue);
    priv->queue = NULL;
  }

  G_OBJECT_CLASS (mai_transaction_manager_parent_class)->dispose (object);
}

static void
mai_transaction_manager_class_init (MaiTransactionManagerClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (MaiTransactionManagerPrivate));

  object_class->dispose = _dispose;
}

static void
mai_transaction_manager_init (MaiTransactionManager *self)
{
  MaiTransactionManagerPrivate *priv = GET_PRIVATE (self);

  priv->queue = g_queue_new ();
}

MaiTransactionManager *
mai_transaction_manager_new (void)
{
  return g_object_new (MAI_TYPE_TRANSACTION_MANAGER, NULL);
}

static void
_transaction_finished_cb (MaiTransaction        *transaction,
                          MaiTransactionManager *self)
{
  MaiTransactionManagerPrivate *priv = GET_PRIVATE (self);

  g_return_if_fail (transaction == g_queue_peek_head (priv->queue));

  g_signal_handlers_disconnect_by_func (transaction,
                                        MAI_CALLBACK_TO_POINTER (_transaction_finished_cb),
                                        self);

  g_queue_pop_head (priv->queue);
  g_object_unref (transaction);

  if (!g_queue_is_empty (priv->queue))
  {
    GError *error = NULL;
    MaiTransaction *next = g_queue_peek_head (priv->queue);
    mai_transaction_run (next, &error);
    if (error)
    {
      g_warning ("%s : %s", G_STRLOC, error->message);
      g_clear_error (&error);
    }
  }
}

gboolean
mai_transaction_manager_queue (MaiTransactionManager   *self,
                               MaiPackage              *package,
                               MaiPackageStatus         target_status,
                               GError                 **error)
{
  MaiTransactionManagerPrivate  *priv = GET_PRIVATE (self);
  MaiTransaction                *transaction;
  gboolean                       ret = FALSE;

  g_return_val_if_fail (MAI_IS_TRANSACTION_MANAGER (self), FALSE);

  transaction = mai_transaction_new (package, target_status);
  g_return_val_if_fail (transaction, FALSE);

  g_queue_push_tail (priv->queue, transaction);
  g_signal_connect (transaction, "finished",
                    G_CALLBACK (_transaction_finished_cb), self);

  if (g_queue_get_length (priv->queue) == 1)
  {
    ret = mai_transaction_run (transaction, error);
  } else {
    /* FIXME create error */
  }

  return ret;
}

static gint
_find_transaction (const MaiTransaction *transaction,
                   const gchar          *package_name)
{
  MaiPackage  *package;
  const gchar *name;

  package = mai_transaction_get_package (transaction);
  name = mai_package_get_package_name (package);

  /* No match if both are NULL. */
  g_return_val_if_fail (name && package_name, 1);

  return g_strcmp0 (name, package_name);
}

gboolean
mai_transaction_manager_cancel_by_name (MaiTransactionManager   *self,
                                        const gchar             *package_name,
                                        GError                 **error)
{
  MaiTransactionManagerPrivate  *priv = GET_PRIVATE (self);
  GList                         *item;
  MaiTransaction                *transaction;
  gboolean                       ret = FALSE;

  g_return_val_if_fail (MAI_IS_TRANSACTION_MANAGER (self), FALSE);
  g_return_val_if_fail (!g_queue_is_empty (priv->queue), FALSE);

  /* Ensure transaction is queued. */
  item = g_queue_find_custom (priv->queue,
                              package_name,
                              (GCompareFunc) _find_transaction);
  g_return_val_if_fail (item, FALSE);

  transaction = MAI_TRANSACTION (item->data);
  if (transaction == g_queue_peek_head (priv->queue))
  {
    /* Transaction is active at the moment and thus can not be cancelled */
    /* FIXME create error */
    g_return_val_if_fail (0, FALSE);
  } else if (!g_queue_is_empty (priv->queue)) {
    /* FIXME make sure transaction is queued */
    g_queue_delete_link (priv->queue, item);
    g_object_unref (transaction);
  } else {
    /* FIXME create error */
  }

  return ret;
}

