
/*
 * moblin-app-installer -- Moblin garage client application.
 * Copyright (c) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Author: Robert Staudinger <robertx.staudinger@intel.com>
 */

#include "mai-transaction.h"

G_DEFINE_TYPE (MaiTransaction, mai_transaction, G_TYPE_OBJECT)

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), MAI_TYPE_TRANSACTION, MaiTransactionPrivate))

enum
{
  PROP_0,

  PROP_PACKAGE,
  PROP_TARGET_STATUS
};

enum
{
  FINISHED,

  LAST_SIGNAL
};

typedef struct
{
  MaiPackage        *package;
  MaiPackageStatus   initial_status;
  MaiPackageStatus   target_status;
  gboolean           is_queued;
  gulong             status_notify_id;
} MaiTransactionPrivate;

static void
mai_transaction_set_target_status (MaiTransaction   *self,
                                   MaiPackageStatus  target_status);

static void
mai_transaction_set_package       (MaiTransaction   *self,
                                   MaiPackage       *package);

static guint _signals[LAST_SIGNAL] = { 0, };

static void
_package_status_notify_cb (MaiPackage     *package,
                           GParamSpec     *pspec,
                           MaiTransaction *self)
{
  MaiPackageStatus status;

  status = mai_package_get_status (package);
  switch (status)
  {
  case MAI_PACKAGE_STATUS_AVAILABLE:
  case MAI_PACKAGE_STATUS_INSTALLED:
  case MAI_PACKAGE_STATUS_ERROR:
    /* Finished. */
    g_signal_emit_by_name (self, "finished");
    break;
  case MAI_PACKAGE_STATUS_QUEUED_FOR_RESOLVE:
  case MAI_PACKAGE_STATUS_RESOLVING:
  case MAI_PACKAGE_STATUS_QUEUED_FOR_INSTALLATION:
  case MAI_PACKAGE_STATUS_INSTALLING:
  case MAI_PACKAGE_STATUS_QUEUED_FOR_REMOVAL:
  case MAI_PACKAGE_STATUS_REMOVING:
    /* Not finished yet. */
    break;
  }
}

static void
_get_property (GObject     *object,
               guint        property_id,
               GValue      *value,
               GParamSpec  *pspec)
{
  switch (property_id)
  {
  case PROP_PACKAGE:
    g_value_set_object (value,
                        mai_transaction_get_package (MAI_TRANSACTION (object)));
    break;
  case PROP_TARGET_STATUS:
    g_value_set_uint (value,
                      mai_transaction_get_target_status (MAI_TRANSACTION (object)));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_set_property (GObject       *object,
               guint          property_id,
               const GValue  *value,
               GParamSpec    *pspec)
{
  switch (property_id)
  {
  case PROP_PACKAGE:
    mai_transaction_set_package (MAI_TRANSACTION (object),
                                 g_value_get_object (value));
    break;
  case PROP_TARGET_STATUS:
    mai_transaction_set_target_status (MAI_TRANSACTION (object),
                                       g_value_get_uint (value));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_dispose (GObject *object)
{
  MaiTransactionPrivate *priv = GET_PRIVATE (object);
  MaiPackageStatus       status;

  if (priv->package)
  {
    if (priv->status_notify_id)
    {
      g_signal_handler_disconnect (priv->package, priv->status_notify_id);
      priv->status_notify_id = 0;
    }

    /* Reset package status if transaction is disposed while queued,
     * i.e. before run. */
    status = mai_package_get_status (priv->package);
    if (MAI_PACKAGE_STATUS_QUEUED_FOR_RESOLVE == status ||
        MAI_PACKAGE_STATUS_QUEUED_FOR_INSTALLATION == status ||
        MAI_PACKAGE_STATUS_QUEUED_FOR_REMOVAL == status)
    {
      mai_package_set_status (priv->package, priv->initial_status);
    } else {
      /* FIXME: try to cancel. */
    }

    g_object_unref (priv->package);
    priv->package = NULL;
  }

  G_OBJECT_CLASS (mai_transaction_parent_class)->dispose (object);
}

static GObject *
_constructor (GType                  type,
              guint                  n_properties,
              GObjectConstructParam *properties)
{
  MaiTransaction *self = (MaiTransaction *)
                            G_OBJECT_CLASS (mai_transaction_parent_class)
                              ->constructor (type, n_properties, properties);
  MaiTransactionPrivate *priv = GET_PRIVATE (self);

  g_return_val_if_fail (priv->package, NULL);

  priv->initial_status = mai_package_get_status (priv->package);

  switch (priv->target_status)
  {
  case MAI_PACKAGE_STATUS_QUEUED_FOR_RESOLVE:
  case MAI_PACKAGE_STATUS_RESOLVING:
    mai_package_set_status (priv->package,
                            MAI_PACKAGE_STATUS_QUEUED_FOR_RESOLVE);
    break;
  case MAI_PACKAGE_STATUS_QUEUED_FOR_INSTALLATION:
  case MAI_PACKAGE_STATUS_INSTALLING:
    mai_package_set_status (priv->package,
                            MAI_PACKAGE_STATUS_QUEUED_FOR_INSTALLATION);
    break;
  case MAI_PACKAGE_STATUS_QUEUED_FOR_REMOVAL:
  case MAI_PACKAGE_STATUS_REMOVING:
    mai_package_set_status (priv->package,
                            MAI_PACKAGE_STATUS_QUEUED_FOR_REMOVAL);
    break;
  case MAI_PACKAGE_STATUS_AVAILABLE:
  case MAI_PACKAGE_STATUS_INSTALLED:
  case MAI_PACKAGE_STATUS_ERROR:
    g_assert_not_reached ();
    return NULL;
  }

  return G_OBJECT (self);
}

static void
mai_transaction_class_init (MaiTransactionClass *klass)
{
  GObjectClass  *object_class = G_OBJECT_CLASS (klass);
  GParamFlags    param_flags;

  g_type_class_add_private (klass, sizeof (MaiTransactionPrivate));

  object_class->constructor = _constructor;
  object_class->get_property = _get_property;
  object_class->set_property = _set_property;
  object_class->dispose = _dispose;

  param_flags = G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY;

  g_object_class_install_property (object_class,
                                   PROP_PACKAGE,
                                   g_param_spec_object ("package",
                                                        "Package",
                                                        "The package",
                                                        MAI_TYPE_PACKAGE,
                                                        param_flags));
  g_object_class_install_property (object_class,
                                   PROP_TARGET_STATUS,
                                   g_param_spec_uint ("target-status",
                                                      "Target status",
                                                      "Transaction target status",
                                                      MAI_PACKAGE_STATUS_RESOLVING,
                                                      MAI_PACKAGE_STATUS_ERROR,
                                                      MAI_PACKAGE_STATUS_RESOLVING,
                                                      param_flags));
  /* Signals. */

  _signals[FINISHED] = g_signal_new ("finished",
                                     G_TYPE_FROM_CLASS (klass),
                                     G_SIGNAL_RUN_LAST,
                                     G_STRUCT_OFFSET (MaiTransactionClass, finished),
                                     NULL, NULL,
                                     g_cclosure_marshal_VOID__VOID,
                                     G_TYPE_NONE, 0);
}

static void
mai_transaction_init (MaiTransaction *self)
{
}

MaiTransaction *
mai_transaction_new (MaiPackage       *package,
                     MaiPackageStatus  target_status)
{
  return g_object_new (MAI_TYPE_TRANSACTION,
                       "package", package,
                       "target-status", target_status,
                       NULL);
}

MaiPackageStatus
mai_transaction_get_target_status (MaiTransaction const *self)
{
  g_return_val_if_fail (MAI_IS_TRANSACTION (self), MAI_PACKAGE_STATUS_ERROR);

  return GET_PRIVATE (self)->target_status;
}

static void
mai_transaction_set_target_status (MaiTransaction   *self,
                                   MaiPackageStatus  target_status)
{
  MaiTransactionPrivate *priv;

  g_return_if_fail (MAI_IS_TRANSACTION (self));
  g_return_if_fail (target_status == MAI_PACKAGE_STATUS_RESOLVING ||
                    target_status == MAI_PACKAGE_STATUS_INSTALLING ||
                    target_status == MAI_PACKAGE_STATUS_REMOVING);

  priv = GET_PRIVATE (self);

  if (target_status != priv->target_status)
  {
    priv->target_status = target_status;
    g_object_notify (G_OBJECT (self), "target-status");
  }
}

MaiPackage *
mai_transaction_get_package (MaiTransaction const *self)
{
  g_return_val_if_fail (MAI_IS_TRANSACTION (self), NULL);

  return GET_PRIVATE (self)->package;
}

static void
mai_transaction_set_package (MaiTransaction *self,
                             MaiPackage     *package)
{
  MaiTransactionPrivate *priv;

  g_return_if_fail (MAI_IS_TRANSACTION (self));
  g_return_if_fail (MAI_IS_PACKAGE (package));

  priv = GET_PRIVATE (self);

  if (package != priv->package)
  {
    if (priv->package)
    {
      g_object_unref (priv->package);
      priv->package = NULL;
    }

    if (package)
    {
      priv->package = g_object_ref (package);
      g_object_notify (G_OBJECT (self), "package");

      priv->status_notify_id = g_signal_connect (priv->package, "notify::status",
                                                 G_CALLBACK (_package_status_notify_cb), self);
    }
  }
}

gboolean
mai_transaction_run (MaiTransaction   *self,
                     GError          **error)
{
  MaiTransactionPrivate *priv = GET_PRIVATE (self);

  switch (priv->target_status)
  {
  case MAI_PACKAGE_STATUS_RESOLVING:
    return mai_package_resolve (priv->package, error);
  case MAI_PACKAGE_STATUS_INSTALLING:
    return mai_package_install (priv->package, error);
  case MAI_PACKAGE_STATUS_REMOVING:
    return mai_package_remove (priv->package, error);
  default:
    g_assert_not_reached ();
  }
}

