
/*
 * moblin-app-installer -- Moblin garage client application.
 * Copyright (c) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Author: Robert Staudinger <robertx.staudinger@intel.com>
 */

#include <glib/gi18n.h>
#include <gtk/gtk.h>

/* FIXME make conditional. */
#include <rpm/rpmlib.h>

#include "mai-model-priv.h"
#include "mai-package.h"
#include "mai-utils.h"
#include "config.h"

#define MAI_PACKAGE_ERROR (mai_package_error_quark ())

static GQuark
mai_package_error_quark (void)
{
  static GQuark _quark = 0;
  if (!_quark)
    _quark = g_quark_from_static_string ("mai-package-error");
  return _quark;
}

#define MAI_GOBJECT_PROPERTY_TRANSACTION(gobject_, statements_)               \
          {                                                                   \
            g_object_freeze_notify (G_OBJECT (gobject_));                     \
            statements_                                                       \
            g_object_thaw_notify (G_OBJECT (gobject_));                       \
          }

/*
 * MaiPackage
 */

G_DEFINE_TYPE (MaiPackage, mai_package, G_TYPE_OBJECT)

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), MAI_TYPE_PACKAGE, MaiPackagePrivate))

enum
{
  PROP_0,

  PROP_PACKAGE_ID,
  PROP_PACKAGE_NAME,
  PROP_NAME,
  PROP_VERSION,
  PROP_DESCRIPTION,
  PROP_LONG_DESCRIPTION,
  PROP_ICON,
  PROP_CATEGORIES,
  PROP_STATUS,
  PROP_PROGRESS,
  PROP_MESSAGE
};

typedef struct {
  /* Properties */
  char                 *package_id;
  gchar                *package_name;
  gchar                *name;
  gchar                *version;
  gchar                *description;
  gchar                *long_description;
  gchar                *icon;
  gchar               **categories;
  MaiPackageStatus      status;
  gint                  progress;
  /* Data */
  gchar                *message;
} MaiPackagePrivate;

static void
mai_package_set_name              (MaiPackage       *self,
                                   const gchar      *name);
static void
mai_package_set_version           (MaiPackage       *self,
                                   const gchar      *version);
static void
mai_package_set_description       (MaiPackage       *self,
                                   const gchar      *description);
static void
mai_package_set_long_description  (MaiPackage       *self,
                                  const gchar       *long_description);
static void
mai_package_set_icon              (MaiPackage       *self,
                                   const gchar      *icon);
static void
mai_package_set_categories        (MaiPackage       *self,
                                   const gchar     **categories);
static void
mai_package_set_progress          (MaiPackage       *self,
                                   gint              percentage);

static const gchar *
mai_package_status_to_text (MaiPackage const *self)
{
  MaiPackagePrivate *priv;

  g_return_val_if_fail (self, NULL);

  priv = GET_PRIVATE (self);

  switch (priv->status)
  {
  case MAI_PACKAGE_STATUS_QUEUED_FOR_RESOLVE:
    return _("Queued for resolve");
  case MAI_PACKAGE_STATUS_RESOLVING:
    return _("Resolving ...");
  case MAI_PACKAGE_STATUS_AVAILABLE:
    return _("Available");
  case MAI_PACKAGE_STATUS_QUEUED_FOR_INSTALLATION:
    return _("Queued for installation");
  case MAI_PACKAGE_STATUS_INSTALLING:
    return _("Installing ...");
  case MAI_PACKAGE_STATUS_INSTALLED:
    return _("Installed");
  case MAI_PACKAGE_STATUS_QUEUED_FOR_REMOVAL:
    return _("Queued for removal");
  case MAI_PACKAGE_STATUS_REMOVING:
    return _("Uninstalling ...");
  case MAI_PACKAGE_STATUS_ERROR:
    if (priv->message)
      return priv->message;
    else
      return _("Not found");
  }

  return NULL;
}

static void
_get_property (GObject    *object,
               guint       property_id,
               GValue     *value,
               GParamSpec *pspec)
{
  switch (property_id)
  {
  case PROP_PACKAGE_ID:
    g_value_set_string (value,
                        mai_package_get_package_id (MAI_PACKAGE (object)));
    break;
  case PROP_PACKAGE_NAME:
    g_value_set_string (value,
                        mai_package_get_package_name (MAI_PACKAGE (object)));
    break;
  case PROP_NAME:
    g_value_set_string (value,
                        mai_package_get_name (MAI_PACKAGE (object)));
    break;
  case PROP_VERSION:
    g_value_set_string (value,
                        mai_package_get_version (MAI_PACKAGE (object)));
    break;
  case PROP_DESCRIPTION:
    g_value_set_string (value,
                        mai_package_get_description (MAI_PACKAGE (object)));
    break;
  case PROP_LONG_DESCRIPTION:
    g_value_set_string (value,
                        mai_package_get_long_description (MAI_PACKAGE (object)));
    break;
  case PROP_ICON:
    g_value_set_string (value,
                        mai_package_get_icon (MAI_PACKAGE (object)));
    break;
  case PROP_CATEGORIES:
    g_value_set_pointer (value,
                         mai_package_get_categories (MAI_PACKAGE (object)));
    break;
  case PROP_STATUS:
    g_value_set_uint (value,
                      mai_package_get_status (MAI_PACKAGE (object)));
    break;
  case PROP_PROGRESS:
    g_value_set_int (value,
                     mai_package_get_progress (MAI_PACKAGE (object)));
    break;
  case PROP_MESSAGE:
    g_value_set_string (value,
                        mai_package_get_message (MAI_PACKAGE (object)));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_set_property (GObject      *object,
               guint         property_id,
               const GValue *value,
               GParamSpec   *pspec)
{
  MaiPackagePrivate *priv = GET_PRIVATE (object);

  switch (property_id)
  {
  case PROP_PACKAGE_ID:
    g_warn_if_fail (priv->package_id == NULL);
    priv->package_id = g_value_dup_string (value);
    break;
  case PROP_PACKAGE_NAME:
    priv->package_name = g_value_dup_string (value);
    break;
  case PROP_NAME:
    mai_package_set_name (MAI_PACKAGE (object),
                          g_value_get_string (value));
    break;
  case PROP_VERSION:
    mai_package_set_version (MAI_PACKAGE (object),
                             g_value_get_string (value));
    break;
  case PROP_DESCRIPTION:
    mai_package_set_description (MAI_PACKAGE (object),
                                 g_value_get_string (value));
    break;
  case PROP_LONG_DESCRIPTION:
    mai_package_set_long_description (MAI_PACKAGE (object),
                                      g_value_get_string (value));
    break;
  case PROP_ICON:
    mai_package_set_icon (MAI_PACKAGE (object),
                          g_value_get_string (value));
    break;
  case PROP_CATEGORIES:
    mai_package_set_categories (MAI_PACKAGE (object),
                                g_value_get_pointer (value));
    break;
  case PROP_STATUS:
    mai_package_set_status (MAI_PACKAGE (object),
                            g_value_get_uint (value));
    break;
  case PROP_PROGRESS:
    mai_package_set_progress (MAI_PACKAGE (object),
                              g_value_get_int (value));
    break;
  case PROP_MESSAGE:
    mai_package_set_message (MAI_PACKAGE (object),
                             g_value_get_string (value));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_dispose (GObject *object)
{
  MaiPackagePrivate *priv = GET_PRIVATE (object);

  if (priv->package_id)
    g_free (priv->package_id), priv->package_id = NULL;

  if (priv->package_name)
    g_free (priv->package_name), priv->package_name = NULL;

  if (priv->name)
    g_free (priv->name), priv->name = NULL;

  if (priv->version)
    g_free (priv->version), priv->version = NULL;

  if (priv->description)
    g_free (priv->description), priv->description = NULL;

  if (priv->long_description)
    g_free (priv->long_description), priv->long_description = NULL;

  if (priv->icon)
    g_free (priv->icon), priv->icon = NULL;

  if (priv->categories)
    g_strfreev (priv->categories), priv->categories = NULL;

  if (priv->message)
    g_free (priv->message), priv->message = NULL;

  G_OBJECT_CLASS (mai_package_parent_class)->dispose (object);
}

static GObject *
_constructor (GType                  type,
              guint                  n_properties,
              GObjectConstructParam *properties)
{
  MaiPackage *self = (MaiPackage *)
                            G_OBJECT_CLASS (mai_package_parent_class)
                              ->constructor (type, n_properties, properties);
  const gchar *icon;

  /* Icon */
  /* FIXME: check the icon exists, otherwise set default too. */
  if (NULL == (icon = mai_package_get_icon (self)))
  {
    /* No icon, use generic one. */
    GtkIconInfo *info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (),
                                                    "application-x-executable",
                                                    48,
                                                    GTK_ICON_LOOKUP_GENERIC_FALLBACK);
    if (info)
    {
      mai_package_set_icon (self, gtk_icon_info_get_filename (info));
      gtk_icon_info_free (info);
    }
  }

  return G_OBJECT (self);
}

static void
mai_package_class_init (MaiPackageClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GParamFlags   param_flags;

  g_type_class_add_private (klass, sizeof (MaiPackagePrivate));

  object_class->constructor = _constructor;
  object_class->get_property = _get_property;
  object_class->set_property = _set_property;
  object_class->dispose = _dispose;

  param_flags = G_PARAM_READWRITE | G_PARAM_CONSTRUCT;

  g_object_class_install_property (object_class,
                                   PROP_PACKAGE_ID,
                                   g_param_spec_string ("package-id", "Package ID",
                                                        "ID of package",
                                                        NULL, G_PARAM_READWRITE));
  g_object_class_install_property (object_class,
                                   PROP_PACKAGE_NAME,
                                   g_param_spec_string ("package-name", "Package name",
                                                        "Name of package",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_NAME,
                                   g_param_spec_string ("name", "Name",
                                                        "Application name",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_VERSION,
                                   g_param_spec_string ("version", "Version",
                                                        "Application version",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_DESCRIPTION,
                                   g_param_spec_string ("description", "Description",
                                                        "Application description",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_LONG_DESCRIPTION,
                                   g_param_spec_string ("long-description",
                                                        "Long description",
                                                        "Long package description",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_ICON,
                                   g_param_spec_string ("icon", "Icon",
                                                        "Application icon",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_CATEGORIES,
                                   g_param_spec_pointer ("categories", "Categories",
                                                         "Application categories",
                                                         param_flags));
  /* FIXME turn into an enum derived type. */
  g_object_class_install_property (object_class,
                                   PROP_STATUS,
                                   g_param_spec_uint ("status", "Status",
                                                      "Package status",
                                                      MAI_PACKAGE_STATUS_RESOLVING,
                                                      MAI_PACKAGE_STATUS_ERROR,
                                                      MAI_PACKAGE_STATUS_RESOLVING,
                                                      param_flags));
  g_object_class_install_property (object_class,
                                   PROP_PROGRESS,
                                   g_param_spec_int ("progress", "Progress",
                                                      "Transaction progress",
                                                      -1, 100, -1,
                                                      param_flags));
  g_object_class_install_property (object_class,
                                   PROP_MESSAGE,
                                   g_param_spec_string ("message",
                                                        "Message",
                                                        "Package status message",
                                                        NULL, param_flags));
}

static void
mai_package_init (MaiPackage *self)
{
}

MaiPackage *
mai_package_new (const gchar  *package,
                 const gchar  *name,
                 const gchar  *version,
                 const gchar  *description,
                 const gchar  *long_description,
                 const gchar  *icon,
                 const gchar **categories)
{
  MaiPackage *self;

  g_return_val_if_fail (package, NULL);

  g_debug ("%s() '%s'", __FUNCTION__, package);

  self = g_object_new (MAI_TYPE_PACKAGE,
                      "package-name",     package,
                      "name",             name,
                      "version",          version,
                      "description",      description,
                      "long-description", long_description,
                      "icon",             icon,
                      "categories",       categories,
                      NULL);

  return self;
}

gchar const *
mai_package_get_package_id (MaiPackage *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->package_id;
}

void
mai_package_set_package_id (MaiPackage    *self,
                            gchar const   *package_id)
{
  MaiPackagePrivate  *priv;
  gchar             **id;

  g_return_if_fail (self);
  g_return_if_fail (package_id);

  priv = GET_PRIVATE (self);
  id = pk_package_ids_from_id (priv->package_id);
  g_return_if_fail (0 == g_strcmp0 (priv->package_name, id[0]));

  if (priv->package_id)
  {
    g_free (priv->package_id);
    priv->package_id = NULL;
  }
  if (package_id)
  {
    priv->package_id = g_strdup (package_id);
  }

  /* No notification because this is considered a private property. */

  mai_package_set_version (self, id[1]);

  g_strfreev (id);
}

const gchar *
mai_package_get_package_name (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->package_name;
}

const gchar *
mai_package_get_name (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->name;
}

static void
mai_package_set_name (MaiPackage  *self,
                      const gchar *name)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);
  g_return_if_fail (name);

  priv = GET_PRIVATE (self);
  if (0 != g_strcmp0 (name, priv->name))
  {
    g_free (priv->name);
    priv->name = g_strdup (name);
    g_object_notify (G_OBJECT (self), "name");
  }
}

const gchar *
mai_package_get_version (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->version;
}

static void
mai_package_set_version (MaiPackage   *self,
                         const gchar  *version)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);

  priv = GET_PRIVATE (self);
  if (0 != g_strcmp0 (version, priv->version))
  {
    g_free (priv->version);
    priv->version = g_strdup (version);
    g_object_notify (G_OBJECT (self), "version");
  }
}

const gchar *
mai_package_get_description (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->description;
}

static void
mai_package_set_description (MaiPackage   *self,
                             const gchar  *description)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);
  g_return_if_fail (description);

  priv = GET_PRIVATE (self);
  if (0 != g_strcmp0 (description, priv->description))
  {
    g_free (priv->description);
    priv->description = g_strdup (description);
    g_object_notify (G_OBJECT (self), "description");
  }
}

const gchar *
mai_package_get_long_description (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->long_description;
}

static void
mai_package_set_long_description (MaiPackage   *self,
                                  const gchar  *long_description)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);
  g_return_if_fail (long_description);

  priv = GET_PRIVATE (self);
  if (0 != g_strcmp0 (long_description, priv->long_description))
  {
    g_free (priv->long_description);
    priv->long_description = g_strdup (long_description);
    g_object_notify (G_OBJECT (self), "long-description");
  }
}

const gchar *
mai_package_get_icon (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->icon;
}

static void
mai_package_set_icon (MaiPackage  *self,
                      const gchar *icon)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);

  priv = GET_PRIVATE (self);
  if (0 != g_strcmp0 (icon, priv->icon))
  {
    g_free (priv->icon);
    priv->icon = g_strdup (icon);
    g_object_notify (G_OBJECT (self), "icon");
  }
}

MaiPackageStatus
mai_package_get_status (MaiPackage const *self)
{
  g_return_val_if_fail (self, MAI_PACKAGE_STATUS_ERROR);

  return GET_PRIVATE (self)->status;
}

void
mai_package_set_status (MaiPackage       *self,
                        MaiPackageStatus  status)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);
  g_return_if_fail (status >= MAI_PACKAGE_STATUS_RESOLVING);
  g_return_if_fail (status <= MAI_PACKAGE_STATUS_ERROR);

  priv = GET_PRIVATE (self);
  if (status != priv->status)
  {
    priv->status = status;
    g_object_notify (G_OBJECT (self), "status");
  }
}

const gchar **
mai_package_get_categories (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return (const gchar **) GET_PRIVATE (self)->categories;
}

static void
mai_package_set_categories (MaiPackage   *self,
                            const gchar **categories)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);

  priv = GET_PRIVATE (self);

  g_strfreev (priv->categories);
  priv->categories = g_strdupv ((gchar **) categories);
  g_object_notify (G_OBJECT (self), "categories");
}

const gchar *
mai_package_get_message (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->message;
}

void
mai_package_set_message (MaiPackage   *self,
                         const gchar  *message)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);

  priv = GET_PRIVATE (self);

  if (0 != g_strcmp0 (priv->message, message))
  {
    g_free (priv->message);
    priv->message = g_strdup (message);
    g_object_notify (G_OBJECT (self), "message");
  }
}

gint
mai_package_get_progress (MaiPackage const *self)
{
  g_return_val_if_fail (self, 0);

  return GET_PRIVATE (self)->progress;
}

static void
mai_package_set_progress (MaiPackage  *self,
                          gint         percentage)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);
  g_warn_if_fail (percentage >= -1);
  g_warn_if_fail (percentage <= 100);

  priv = GET_PRIVATE (self);
  percentage = CLAMP (percentage, -1, 100);
  if (percentage != priv->progress)
  {
    priv->progress = percentage;
    g_object_notify (G_OBJECT (self), "progress");
  }
}

/*
 * returns > 0 if self is newer
 *           0 if equal
 *         < 0 if version is newer
 */
gint
mai_package_compare_version (MaiPackage const  *self,
                             const gchar       *version)
{
  MaiPackagePrivate *priv;

  g_return_val_if_fail (self, 0);

  priv = GET_PRIVATE (self);
  g_return_val_if_fail (priv->version && version, 0);

  if (NULL == priv->package_id)
    return -1;

  /* FIXME support deb. */
  return rpmvercmp (priv->version, version);
}

static void
_progress_cb (PkProgress      *progress,
              PkProgressType   type,
              MaiPackage      *self)
{
  switch (type)
  {
	case PK_PROGRESS_TYPE_PERCENTAGE:
    {
      gint percentage;
      g_object_get (progress, "percentage", &percentage, NULL);
      mai_package_set_progress (self, percentage);
    }
    break;
	case PK_PROGRESS_TYPE_STATUS:
    {
      PkStatusEnum status;
      g_object_get (progress, "status", &status, NULL);
      mai_package_set_message (self, pk_status_enum_to_text (status));
    }
    break;
  default:
    ;
  }
}

static gboolean
finish_transaction (MaiPackage  *self,
                    PkResults   *result,
                    GError      *error)
{
  PkExitEnum exit;

  if (error)
  {
    mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
    mai_package_set_message (self, error->message);
    g_warning ("%s : %s", G_STRLOC, error->message);
    return FALSE;
  }

  exit = pk_results_get_exit_code (result);
  if (exit != PK_EXIT_ENUM_SUCCESS)
  {
    mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
    mai_package_set_message (self, error->message);
    g_warning ("%s : %s", G_STRLOC, pk_network_enum_to_text (exit));
    // TODO resolve if not already tried resolving
    return FALSE;
  }

  return TRUE;
}

static void
_resolve_finished_cb (PkClient      *client,
                      GAsyncResult  *res,
                      MaiPackage    *self)
{
  MaiPackagePrivate *priv = GET_PRIVATE (self);
  PkResults         *result;
  GError            *error = NULL;

  g_debug ("%s()", __FUNCTION__);

  result = pk_client_generic_finish (client, res, &error);
  if (finish_transaction (self, result, error) &&
      priv->status == MAI_PACKAGE_STATUS_RESOLVING)
  {
    GPtrArray *pkgs = pk_results_get_package_array (result);
    gint i;
    PkPackage *pkg = NULL;
    for (i = 0, pkg = g_ptr_array_index (pkgs, i);
         pkg;
         pkg = g_ptr_array_index (pkgs, i), i++)
    {
      const gchar *package_id = pk_package_get_id (pkg);
      gchar **id = pk_package_id_split (package_id);

      if (!priv->version ||
          rpmvercmp (id[PK_PACKAGE_ID_VERSION], priv->version) > 0)
      {
        PkInfoEnum package_info;

        mai_package_set_package_id (self, package_id);

        /* Update package status. */
        g_object_get (pkg, "info", &package_info, NULL);
        if (package_info == PK_INFO_ENUM_INSTALLED)
          mai_package_set_status (self, MAI_PACKAGE_STATUS_INSTALLED);
        else if (package_info == PK_INFO_ENUM_AVAILABLE)
        {
          mai_package_set_status (self, MAI_PACKAGE_STATUS_AVAILABLE);
#if 0
          TODO keep network state in the model.
          /* Set available if we have network connection. */
          PkControl *control = mai_model_get_control ();
          GError *error = NULL;
          PkNetworkEnum network_state = pk_control_get_network_state (control, &error);
          if (error)
          {
            mai_package_set_message (self, error->message);
            mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
            g_clear_error (&error);
          } else {
            g_debug ("    network: '%s'", pk_network_enum_to_text (network_state));
            if (network_state == PK_NETWORK_ENUM_ONLINE ||
                network_state == PK_NETWORK_ENUM_WIRED ||
                network_state == PK_NETWORK_ENUM_WIFI ||
                network_state == PK_NETWORK_ENUM_MOBILE)
            {
              mai_package_set_status (self, MAI_PACKAGE_STATUS_AVAILABLE);
            } else {
              gchar *message = g_strdup_printf ("Network is '%s'",
                                                pk_network_enum_to_text (network_state));
              mai_package_set_message (self, message);
              g_free (message);
              mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
            }
          }
#endif
        }
      }
      g_strfreev (id);
    }
    g_ptr_array_unref (pkgs);
  } else {
    gchar *message = g_strdup_printf ("Invalid package status '%s'",
                                      mai_package_status_to_text (self));
    mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
    mai_package_set_message (self, message);
    g_free (message);
  }

  g_object_unref (result);
  g_clear_error (&error);
  mai_package_set_progress (self, -1);
}

gboolean
mai_package_resolve (MaiPackage  *self,
                     GError     **error)
{
  MaiPackagePrivate *priv = GET_PRIVATE (self);
  PkClient          *client;
  gchar const       *values[2] = { NULL, };

  g_return_val_if_fail (MAI_IS_PACKAGE (self), FALSE);

  MAI_DEBUG ("'%s'", priv->package_name);

  MAI_GOBJECT_PROPERTY_TRANSACTION (self,
    mai_package_set_status (self, MAI_PACKAGE_STATUS_RESOLVING);
    mai_package_set_progress (self, 0);
  );

  client = pk_client_new ();
  values[0] = priv->package_name;
  pk_client_resolve_async (client, PK_FILTER_ENUM_NONE, (gchar **) values, NULL,
                           (PkProgressCallback) _progress_cb, self,
                           (GAsyncReadyCallback) _resolve_finished_cb, self);
  g_object_unref (client);

  return TRUE;
}

static void
_install_finished_cb (PkClient      *client,
                      GAsyncResult  *res,
                      MaiPackage    *self)
{
  MaiPackagePrivate *priv = GET_PRIVATE (self);
  PkResults         *result;
  GError            *error = NULL;

  g_debug ("%s()", __FUNCTION__);

  result = pk_client_generic_finish (client, res, &error);
  if (finish_transaction (self, result, error) &&
      priv->status == MAI_PACKAGE_STATUS_INSTALLING)
  {
    priv->status = MAI_PACKAGE_STATUS_INSTALLED;
    /* Resolve again to get consistent package-id across backends. */
    mai_package_resolve (self, NULL);
  } else {
    gchar *message = g_strdup_printf ("Invalid package status '%s'",
                                      mai_package_status_to_text (self));
    mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
    mai_package_set_message (self, message);
    g_free (message);
  }

  g_object_unref (result);
  mai_package_set_progress (self, -1);
}

gboolean
mai_package_install (MaiPackage  *self,
                     GError     **error)
{
  MaiPackagePrivate *priv = GET_PRIVATE (self);
  PkClient          *client;
  gchar const       *package_ids[2] = { NULL, };

  g_return_val_if_fail (MAI_IS_PACKAGE (self), FALSE);
  g_return_val_if_fail (priv->package_id, FALSE);

  MAI_DEBUG ("'%s'", priv->package_id);

  if (priv->status != MAI_PACKAGE_STATUS_QUEUED_FOR_INSTALLATION)
  {
    if (error)
      *error = g_error_new (MAI_PACKAGE_ERROR,
                            MAI_PACKAGE_ERROR_OPERATION_UNAVAILABLE,
                            "Package status is '%s', and therefore can not be installed",
                            mai_package_status_to_text (self));
    return FALSE;
  }

  MAI_GOBJECT_PROPERTY_TRANSACTION (self,
    mai_package_set_status (self, MAI_PACKAGE_STATUS_INSTALLING);
    mai_package_set_progress (self, 0);
  );

  client = pk_client_new ();
  package_ids[0] = priv->package_id;
  pk_client_install_packages_async (client, TRUE, (gchar **) package_ids, NULL,
                                    (PkProgressCallback) _progress_cb, self,
                                    (GAsyncReadyCallback) _install_finished_cb, self);
  g_object_unref (client);

  return TRUE;
}

static void
_remove_finished_cb (PkClient      *client,
                     GAsyncResult  *res,
                     MaiPackage    *self)
{
  MaiPackagePrivate *priv = GET_PRIVATE (self);
  PkResults         *result;
  GError            *error = NULL;

  g_debug ("%s()", __FUNCTION__);

  result = pk_client_generic_finish (client, res, &error);
  if (finish_transaction (self, result, error) &&
      priv->status == MAI_PACKAGE_STATUS_REMOVING)
  {
    priv->status = MAI_PACKAGE_STATUS_AVAILABLE;
    /* Resolve again to get consistent package-id across backends. */
    mai_package_resolve (self, NULL);
  } else {
    gchar *message = g_strdup_printf ("Invalid package status '%s'",
                                      mai_package_status_to_text (self));
    mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
    mai_package_set_message (self, message);
    g_free (message);
  }

  g_object_unref (result);
  mai_package_set_progress (self, -1);
}

gboolean
mai_package_remove (MaiPackage  *self,
                    GError     **error)
{
  MaiPackagePrivate *priv = GET_PRIVATE (self);
  PkClient          *client;
  gchar const       *package_ids[2] = { 0, };

  g_return_val_if_fail (MAI_IS_PACKAGE (self), FALSE);
  g_return_val_if_fail (priv->package_id, FALSE);

  MAI_DEBUG ("'%s'", priv->package_id);

  if (priv->status != MAI_PACKAGE_STATUS_QUEUED_FOR_REMOVAL)
  {
    if (error)
      *error = g_error_new (MAI_PACKAGE_ERROR,
                            MAI_PACKAGE_ERROR_OPERATION_UNAVAILABLE,
                            "Package status is '%s', and therefore can not be uninstalled",
                            mai_package_status_to_text (self));
    return FALSE;
  }

  MAI_GOBJECT_PROPERTY_TRANSACTION (self,
    mai_package_set_status (self, MAI_PACKAGE_STATUS_REMOVING);
    mai_package_set_progress (self, 0);
  );

  client = pk_client_new ();
  package_ids[0] = priv->package_id;
  pk_client_remove_packages_async (client, (gchar **) package_ids, TRUE, TRUE, NULL,
                                    (PkProgressCallback) _progress_cb, self,
                                    (GAsyncReadyCallback) _remove_finished_cb, self);
  g_object_unref (client);

  return TRUE;
}
