
/*
 * moblin-app-installer -- Moblin garage client application.
 * Copyright (c) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Author: Robert Staudinger <robertx.staudinger@intel.com>
 */

#include <glib/gi18n.h>
#define I_KNOW_THE_PACKAGEKIT_GLIB2_API_IS_SUBJECT_TO_CHANGE
#include <packagekit-glib2/packagekit.h>
#include "mai-resolver.h"
#include "mai-package.h"
#include "mai-utils.h"

G_DEFINE_TYPE (MaiResolver, mai_resolver, G_TYPE_OBJECT)

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), MAI_TYPE_RESOLVER, MaiResolverPrivate))

enum
{
  PROP_0,

  PROP_PROGRESS
};

typedef struct {
  /* Properties */
  MaiPackageStatus   status;
  gint               progress;
  /* Data */
  GHashTable  *packages;
  gboolean     have_network;
  gchar       *message;
} MaiResolverPrivate;

static void
mai_resolver_set_progress (MaiResolver        *self,
                           gint                percentage);

static void
_get_property (GObject    *object,
               guint       property_id,
               GValue     *value,
               GParamSpec *pspec)
{
  switch (property_id)
  {
  case PROP_PROGRESS:
    g_value_set_int (value,
                     mai_resolver_get_progress (MAI_RESOLVER (object)));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_set_property (GObject      *object,
               guint         property_id,
               const GValue *value,
               GParamSpec   *pspec)
{
  switch (property_id)
  {
  case PROP_PROGRESS:
    mai_resolver_set_progress (MAI_RESOLVER (object),
                            g_value_get_int (value));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_dispose (GObject *object)
{
  MaiResolverPrivate *priv = GET_PRIVATE (object);

  if (priv->packages)
  {
    g_hash_table_destroy (priv->packages);
    priv->packages = NULL;
  }

  if (priv->message)
  {
    g_free (priv->message);
    priv->message = NULL;
  }

  G_OBJECT_CLASS (mai_resolver_parent_class)->dispose (object);
}

static void
mai_resolver_class_init (MaiResolverClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GParamFlags   param_flags;

  g_type_class_add_private (klass, sizeof (MaiResolverPrivate));

  object_class->get_property = _get_property;
  object_class->set_property = _set_property;
  object_class->dispose = _dispose;

  param_flags = G_PARAM_READWRITE | G_PARAM_CONSTRUCT;

  g_object_class_install_property (object_class,
                                   PROP_PROGRESS,
                                   g_param_spec_int ("progress", "Progress",
                                                      "Transaction progress",
                                                      -1, 100, -1,
                                                      param_flags));
}

static void
mai_resolver_init (MaiResolver *self)
{
  MaiResolverPrivate *priv = GET_PRIVATE (self);

  priv->packages = g_hash_table_new (g_str_hash, g_str_equal);
}

MaiResolver*
mai_resolver_new (void)
{
  return g_object_new (MAI_TYPE_RESOLVER, NULL);
}

void
mai_resolver_add_package (MaiResolver *self,
                          MaiPackage  *package)
{
  MaiResolverPrivate *priv;

  g_return_if_fail (self);
  g_return_if_fail (package);

  priv = GET_PRIVATE (self);

  g_hash_table_insert (priv->packages,
                       (gpointer) mai_package_get_package_name (package),
                       package);
}

static void
_progress_cb (PkProgress      *progress,
              PkProgressType   type,
              MaiResolver     *self)
{
  switch (type)
  {
	case PK_PROGRESS_TYPE_PERCENTAGE:
    {
      gint percentage;
      g_object_get (progress, "percentage", &percentage, NULL);
      mai_resolver_set_progress (self, CLAMP (percentage, 0, 100));
    }
    break;
  default:
    ;
  }
}

static gboolean
finish_transaction (MaiResolver *self,
                    PkResults   *result,
                    GError      *error)
{
  PkExitEnum exit;

  if (error)
  {
    g_critical ("%s : %s", G_STRLOC, error->message);
    return FALSE;
  }

  exit = pk_results_get_exit_code (result);
  if (exit != PK_EXIT_ENUM_SUCCESS)
  {
    g_critical ("%s : %s", G_STRLOC, pk_network_enum_to_text (exit));
    return FALSE;
  }

  return TRUE;
}

static void
_resolve_finished_cb (PkClient      *client,
                      GAsyncResult  *res,
                      MaiResolver   *self)
{
  MaiResolverPrivate *priv = GET_PRIVATE (self);
  PkResults         *result;
  GError            *error = NULL;

  g_debug ("%s()", __FUNCTION__);

  result = pk_client_generic_finish (client, res, &error);
  if (finish_transaction (self, result, error) &&
      priv->status == MAI_PACKAGE_STATUS_RESOLVING)
  {
    GPtrArray *pkgs = pk_results_get_package_array (result);
    gint i;
    PkPackage *pkg = NULL;
    for (i = 0, pkg = g_ptr_array_index (pkgs, i);
         pkg;
         pkg = g_ptr_array_index (pkgs, i), i++)
    {
      const gchar *package_id = pk_package_get_id (pkg);
      gchar **id = pk_package_id_split (package_id);

      MaiPackage *package = g_hash_table_lookup (priv->packages,
                                                 id[PK_PACKAGE_ID_NAME]);
      if (package &&
          mai_package_compare_version (package, id[PK_PACKAGE_ID_VERSION]) < 0)
      {
        mai_package_set_package_id (package, package_id);
#if 0
        // TODO
        if (priv->have_network)
        {
          if (pkg->info == PK_INFO_ENUM_INSTALLED)
            mai_package_set_status (package, MAI_PACKAGE_STATUS_INSTALLED);
          else if (pkg->info == PK_INFO_ENUM_AVAILABLE)
            mai_package_set_status (package, MAI_PACKAGE_STATUS_AVAILABLE);
        } else {
          mai_package_set_message (package, priv->message);
          mai_package_set_status (package, MAI_PACKAGE_STATUS_ERROR);
        }
#endif
      }
      g_strfreev (id);
    }
    g_ptr_array_unref (pkgs);
  }
  g_clear_error (&error);
  mai_resolver_set_progress (self, -1);
}

gboolean
mai_resolver_resolve (MaiResolver  *self,
                      GError      **error)
{
  MaiResolverPrivate *priv = GET_PRIVATE (self);
  PkClient           *client;
  const gchar        *packages[g_hash_table_size (priv->packages) + 1];
  const gchar        *package;
  GHashTableIter      iter;
  guint               i = 0;

  /* Check for network connection. */
priv->have_network = TRUE;
#if 0
  priv->have_network = FALSE;
  if (g_getenv ("MAI_OFFLINE")) {
    priv->message = g_strdup (_("Offline mode"));
  } else {
    PkControl *control = pk_control_new ();
    GError *network_err = NULL;
    PkNetworkEnum network_state = pk_control_get_network_state (control, &network_err);
    g_object_unref (control);
    if (network_err)
    {
      if (error)
        *error = g_error_copy (network_err);
      return FALSE;
    } else if (network_state == PK_NETWORK_ENUM_OFFLINE) {
      priv->message = g_strdup (_("Network is offline"));
    } else {
      priv->have_network = TRUE;
    }
  }
#endif

  /* use /var/log/rpmpkgs if available */
  //return pk_client_get_packages (priv->client, PK_FILTER_ENUM_NONE, error);
  g_hash_table_iter_init (&iter, priv->packages);
  while (g_hash_table_iter_next (&iter, (gpointer *) &package, NULL))
  {
    packages[i] = package;
    i++;
  }
  packages[i] = NULL;

  client = pk_client_new ();
  pk_client_resolve_async (client, PK_FILTER_ENUM_NONE,
                           (gchar **) packages, NULL,
                           (PkProgressCallback) _progress_cb, self,
                           (GAsyncReadyCallback) _resolve_finished_cb, self);
  g_object_unref (client);

  return TRUE;
}

gint
mai_resolver_get_progress (MaiResolver const *self)
{
  g_return_val_if_fail (self, 0);

  return GET_PRIVATE (self)->progress;
}

static void
mai_resolver_set_progress (MaiResolver  *self,
                           gint          percentage)
{
  MaiResolverPrivate *priv;

  g_return_if_fail (self);
  g_warn_if_fail (percentage >= -1);
  g_warn_if_fail (percentage <= 100);

  priv = GET_PRIVATE (self);
  percentage = CLAMP (percentage, -1, 100);
  if (percentage != priv->progress)
  {
    priv->progress = percentage;
    g_object_notify (G_OBJECT (self), "progress");
  }
}

